# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from ast import literal_eval
from collections.abc import Iterable
from typing import Any

from . import grammars
from .builder import ModelBuilderSemantics
from .contexts import ParseContext
from .exceptions import FailedSemantics
from .leftrec import mark_left_recursion
from .util import eval_escapes, re, warning
from .util.abctools import flatten


class TatSuGrammarSemantics(ModelBuilderSemantics):
    def __init__(self, name: str | None = None, context: ParseContext | None = None):
        super().__init__(
            basetype=grammars.Model,
            constructors=grammars.Model.classes(),  # ty:ignore[invalid-argument-type]
        )
        self.name = name
        self.context = context
        self.rulemap: dict[str, grammars.Rule] = {}

    def set_context(self, context: ParseContext):
        self.context = context

    @classmethod
    def _validate_literal(cls, ast: Any):
        try:
            literal_eval(repr(str(ast)))
        except SyntaxError as e:
            raise FailedSemantics('literal string error: ' + str(e)) from e

    @classmethod
    def _validate_pattern(cls, ast: Any):
        cls._validate_literal(ast)
        try:
            re.compile(str(ast))
        except (TypeError, re.error) as e:
            raise FailedSemantics('pattern error: ' + str(e)) from e

    def EMPTYLINE(self, ast: Any, *args) -> Any:
        return ast

    def token(self, ast: str, *args: Any) -> grammars.Token:
        token = ast
        if not token:
            raise FailedSemantics('empty token')
        literal_eval(repr(token))
        return grammars.Token(ast=token)

    def pattern(self, ast: str, *args) -> grammars.Pattern:
        pattern = ast
        self._validate_literal(pattern)
        return grammars.Pattern(ast=pattern)

    def regexes(self, ast: Iterable[str], *args) -> Iterable[str]:
        pattern = ''.join(ast)
        self._validate_pattern(pattern)
        return ast

    def regex(self, ast: str, *args) -> str:
        pattern = ast
        self._validate_pattern(pattern)
        return pattern

    def string(self, ast):
        return eval_escapes(ast)

    def hex(self, ast):
        return int(ast, 16)

    def float(self, ast):
        return float(ast)

    def int(self, ast):
        return int(ast)

    def null(self, ast):
        return None

    def cut_deprecated(self, ast, *args):
        warning('The use of >> for cut is deprecated. Use the ~ symbol instead.')
        return grammars.Cut()

    def override_single_deprecated(self, ast, *args):
        warning('The use of @ for override is deprecated. Use @: instead')
        return grammars.Override(ast)

    def sequence(self, ast, *args):
        # if isinstance(ast, list | tuple):
        #     seq = ast
        # else:
        #     seq = ast.sequence
        seq = ast
        assert isinstance(seq, list), str(seq)
        if len(seq) == 1:
            return seq[0]
        return grammars.Sequence(ast=ast)

    def choice(self, ast):
        return grammars.Choice(ast=ast)

    def new_name(self, name):
        if name in self.rulemap:
            raise FailedSemantics(f'rule "{name!s}" already defined')
        return name

    def known_name(self, name) -> str:
        if name not in self.rulemap:
            raise FailedSemantics(f'rule "{name!s}" not yet defined')
        return name

    def boolean(self, ast):
        return str(ast).lower() in {'true', 'yes', 'ok', '1'}

    def rule(self, ast, *args):
        decorators = ast.decorators
        name = ast.name
        base = ast.base
        params = ast.params
        kwparams = dict(ast.kwparams) if ast.kwparams else {}

        if 'override' not in decorators and name in self.rulemap:
            self.new_name(name)
        elif 'override' in decorators:
            self.known_name(name)

        if not base:
            rule = grammars.Rule(
                ast=ast,
                name=name,
                params=params,
                kwparams=kwparams,
                decorators=decorators,
            )
        else:
            self.known_name(base)
            baserule = self.rulemap[base]
            rule = grammars.BasedRule(
                ast=ast,
                name=name,
                baserule=baserule,
                params=params,
                kwparams=kwparams,
                decorators=decorators,
            )

        self.rulemap[name] = rule
        return rule

    def rule_include(self, ast, *args):
        name = str(ast)
        self.known_name(name)

        rule = self.rulemap[name]
        return grammars.RuleInclude(ast=ast, rule=rule)

    def grammar(self, ast, *args):
        directives = {d.name: d.value for d in flatten(ast.directives)}
        for value in directives.values():
            literal_eval(repr(value))
        keywords = list(flatten(ast.keywords)) or []

        if directives.get('whitespace') in {'None', 'False'}:
            # NOTE: use '' because None will _not_ override defaults in configuration
            directives['whitespace'] = ''

        name = self.name or directives.get('grammar')
        grammar = grammars.Grammar(
            name,
            list(self.rulemap.values()),
            directives=directives,
            keywords=keywords,
        )
        if grammar.config.left_recursion:
            mark_left_recursion(grammar)
        return grammar



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
Parse and translate a TatSu grammar into a Python parser for
the described language.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

from .. import grammars
from ..builder import BuilderConfig, Constructor, ModelBuilderSemantics, TypeContainer
from ..exceptions import ParseException
from ..infos import ParserConfig
from ..ngcodegen.walkgen_model import modelgen
from ..ngcodegen.walkgen_parser import pythongen
from ..objectmodel import Node
from ..parser import TatSuParserGenerator
from ..tokenizing import Tokenizer
from ..util.string import hashsha

__all__ = [
    'compile',
    'gencode',
    'genmodel',
    'grammars',
    'modelgen',
    'parse',
    'pythongen',
    'to_python_model',
    'to_python_sourcecode',
]

__compiled_grammar_cache = {}


def compile(
    grammar: str | Tokenizer,
    name: str | None = None,
    *,
    config: ParserConfig | None = None,
    filename: str | None = None,
    basetype: type | None = None,
    semantics: Any = None,
    asmodel: bool = False,
    builderconfig: BuilderConfig | None = None,
    synthok: bool = True,
    typedefs: list[TypeContainer] | None = None,
    constructors: list[Constructor] | None = None,
    **settings: Any,
) -> grammars.Grammar:
    # check parameters
    ParserConfig.new(
        config=config,
        semantics=semantics,
        name=name,
        filename=filename,
        **settings,
    )
    if isinstance(semantics, type):
        raise TypeError(
            f'semantics must be an object instance or None, not class {semantics!r}',
        )
    cache = __compiled_grammar_cache

    key = (name, hashsha(grammar), id(semantics))
    if key in cache:
        model = cache[key]
    else:
        gen = TatSuParserGenerator(name, **settings)
        model = cache[key] = gen.parse(grammar, **settings)

    asmodel = not semantics and (
        asmodel
        or isinstance(builderconfig, BuilderConfig)
        or basetype is not None
        or typedefs is not None
        or constructors is not None
    )
    if semantics is not None:
        model.semantics = semantics
    elif asmodel:
        # HACK: cheating, but necessary for bw-compatibility
        builderconfig = BuilderConfig.new(
            config=builderconfig,
            synthok=synthok,
            basetype=basetype,
            typedefs=typedefs,
            constructors=constructors,
        )
        model.semantics = ModelBuilderSemantics(config=builderconfig)

    return model


def parse(
    grammar: str,
    text: str,
    /,
    *,
    config: ParserConfig | None = None,
    start: str | None = None,
    name: str | None = None,
    filename: str | None = None,
    semantics: Any | None = None,
    asmodel: bool = False,
    builderconfig: BuilderConfig | None = None,
    basetype: type | None = None,
    synthok: bool = True,
    typedefs: list[TypeContainer] | None = None,
    constructors: list[Constructor] | None = None,
    **settings: Any,
):
    config = ParserConfig.new(
        config=config,
        start=start,
        name=name,
        filename=filename,
        semantics=semantics,
        **settings,
    )
    model = compile(grammar, config=config, asmodel=asmodel)
    config.semantics = semantics or model.semantics

    asmodel = not config.semantics and (
        asmodel
        or isinstance(builderconfig, BuilderConfig)
        or basetype is not None
        or typedefs is not None
        or constructors is not None
    )
    if asmodel:
        builderconfig = BuilderConfig.new(
            config=builderconfig,
            synthok=synthok,
            basetype=basetype,
            typedefs=typedefs,
            constructors=constructors,
        )
        config.semantics = ModelBuilderSemantics(config=builderconfig)
    return model.parse(text, start=start, semantics=semantics, config=config)


def to_python_sourcecode(
    grammar: str,
    /,
    *,
    name: str | None = None,
    filename: str | None = None,
    config: ParserConfig | None = None,
    **settings: Any,
):
    config = ParserConfig.new(config=config, name=name, filename=filename, **settings)
    model = compile(grammar, name=name, filename=filename, config=config)
    return pythongen(model)


def to_python_model(
    grammar: str,
    /,
    *,
    name: str | None = None,
    filename: str | None = None,
    basetype: type = Node,
    config: ParserConfig | None = None,
    **settings: Any,
):
    config = ParserConfig.new(config=config, name=name, filename=filename, **settings)
    model = compile(grammar, name=name, filename=filename, config=config)
    return modelgen(model, basetype=basetype)


# for backwards compatibility. Use `compile()` instead
def genmodel(
    *,
    name: str | None = None,
    grammar: str | None = None,
    semantics: type | None = None,
    config: ParserConfig | None = None,
    **settings: Any,
):
    if grammar is None:
        raise ParseException('grammar is None')

    return compile(grammar, name=name, semantics=semantics, config=config, **settings)


def gencode(
    *,
    name: str | None = None,
    grammar: str,
    trace: bool = False,
    filename: str | None = None,
    codegen: Callable = pythongen,
    config: ParserConfig | None = None,
    **settings: Any,
):
    model = compile(
        grammar,
        name=name,
        filename=filename,
        trace=trace,
        config=config,
        **settings,
    )
    return codegen(model)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import builtins
import types
from collections.abc import Callable, Mapping
from dataclasses import dataclass, field
from typing import Any

from .objectmodel import Node
from .synth import synthesize
from .util.configs import Config
from .util.deprecate import deprecated_params
from .util.string import mangle
from .util.typing import Constructor, TypeContainer, boundcall, least_upper_bound_type


class TypeResolutionError(TypeError):
    """Raised when a constructor for a node type cannot be found or synthesized"""

    pass


def types_defined_in(container: TypeContainer) -> list[type]:
    return ModelBuilder.types_defined_in(container)


@dataclass
class BuilderConfig(Config):
    basetype: type = Node
    synthok: bool = True
    typedefs: list[TypeContainer] = field(default_factory=list)
    constructors: list[Constructor] = field(default_factory=list)


class ModelBuilder:
    """Intended as a semantic action for parsing. A ModelBuildercreates
    nodes using the class name given as first parameter to a grammar
    rule, and synthesizes the class/type if it's not known.
    """

    def __init__(
        self,
        config: BuilderConfig | None = None,
        synthok: bool = True,
        basetype: type | None = None,
        typedefs: list[TypeContainer] | None = None,
        constructors: list[Constructor] | None = None,
    ) -> None:

        config = BuilderConfig.new(
            config=config,
            basetype=basetype,
            synthok=synthok,
            typedefs=typedefs,
            constructors=constructors,
        )
        self._constructor_registry: dict[str, Constructor] = {}

        config = self._typedefs_to_constructors(config)
        config = self._narrow_basetype(config)
        self._register_constructors(config)

        self.config = config
        # HACK!
        self.config = self.config.override(synthok=not constructors)

    def instanceof(
        self, typename, /, known: dict[str, Any], *args: Any, **kwargs: Any,
    ) -> Any:
        return self._instanceof(typename, known, *args, **kwargs)

    def _instanceof(
        self,
        typename: str,
        known: dict[str, Any],
        /,
        *args: Any,
        base: type | None = None,
        **kwargs: Any,
    ) -> Any:
        constructor = self._get_constructor(typename, base=base)
        return boundcall(constructor, known, *args, **kwargs)

    def safe_context(self, *other: Mapping[str, Any]) -> Mapping[str, Any]:
        context: dict[str, Any] = {}
        for ctx in other:
            context |= ctx

        return {name: value for name, value in context.items() if callable(value)}

    @staticmethod
    def types_defined_in(container: TypeContainer, /) -> list[type]:
        contents: dict[str, Any] = {}
        if isinstance(container, types.ModuleType):
            contents.update(vars(container))
        elif isinstance(container, Mapping):
            contents.update(container)  # ty:ignore[no-matching-overload]
        elif hasattr(container, '__dict__'):
            contents.update(vars(container))
        else:
            return []

        type_list = [t for t in contents.values() if isinstance(t, type)]
        name = getattr(container, '__module__', None) or getattr(
            container, '__name__', None,
        )
        if name is None:
            return type_list
        return [t for t in type_list if t.__module__ == name]

    def _funname(self, obj: Any) -> str | None:
        name = getattr(obj, '__name__', None)
        if not name and callable(obj):
            name = type(obj).__name__
        return name

    def _typedefs_to_constructors(self, config: BuilderConfig) -> BuilderConfig:
        if not config.typedefs:
            return config

        contained = []
        for container in config.typedefs:
            contained += self.types_defined_in(container)

        all_constructors = [*config.constructors, *contained]
        return config.override(constructors=all_constructors)

    def _narrow_basetype(self, config: BuilderConfig) -> BuilderConfig:
        if config.basetype and config.basetype is not object:
            return config
        basetype = least_upper_bound_type(config.constructors)
        return config.override(basetype=basetype)

    def _register_constructors(self, config: BuilderConfig) -> None:
        for t in config.constructors:
            if not callable(t):
                raise TypeResolutionError(
                    f'Expected callable in constructors, got: {type(t)!r}',
                )

            name = self._funname(t)
            if not name:
                raise TypeResolutionError(
                    f'Expected __name__ in constructor, got: {type(t)!r}',
                )

            # NOTE: this allows for standalone functions as constructors
            self._register_constructor(t)

    def _register_constructor(self, constructor: Constructor) -> Constructor:
        name = self._funname(constructor)
        if not name:
            return constructor

        existing = self._constructor_registry.get(name)
        if existing and existing is not constructor:
            raise TypeResolutionError(
                f"Conflict for constructor name {name!r}: "
                f"attempted to register {constructor!r}, "
                f"but {existing!r} is already registered.",
            )

        self._constructor_registry[name] = constructor
        return constructor

    def _find_existing_constructor(self, typename: str) -> Constructor | None:
        return self._constructor_registry.get(typename) or vars(builtins).get(typename)

    def _get_constructor(
        self, typename: str, base: type | None, **args: Any,
    ) -> Constructor:
        if constructor := self._find_existing_constructor(typename):
            return constructor

        if not self.config.synthok:
            synthok = bool(self.config.synthok)
            raise TypeResolutionError(
                f'Could not find constructor for type {typename!r}, and {synthok=} ',
            )

        if base is None:
            constructor = synthesize(typename, (), **args)
        else:
            constructor = synthesize(typename, (base,), **args)

        return self._register_constructor(constructor)


class ModelBuilderSemantics:
    @deprecated_params(base_type='basetype', types='constructors', context=None)
    def __init__(
        self,
        config: BuilderConfig | None = None,
        synthok: bool = True,
        basetype: type | None = None,
        typedefs: list[TypeContainer] | None = None,
        constructors: list[Constructor] | None = None,
        # deprecations
        context: Any | None = None,
        base_type: type | None = None,
        types: list[Callable] | None = None,  # for bw compatibility
    ) -> None:
        config = BuilderConfig.new(
            config=config,
            basetype=basetype,
            synthok=synthok,
            typedefs=typedefs,
            constructors=constructors,
        )
        # handle deeprecations
        assert context is None
        if isinstance(base_type, type):
            config = config.override(basetype=base_type)
        if types is not None:
            constructors = [*(config.constructors or []), *types]
            config = config.override(constructors=constructors)

        self.config = config
        self._builder = ModelBuilder(
            config=config,
            synthok=synthok,
            basetype=basetype,
            typedefs=typedefs,
            constructors=constructors,
        )

    def _default(self, ast: Any, *args: Any, **kwargs: Any) -> Any:
        if not args:
            return ast

        typespec = [mangle(s) for s in args[0].split('::')]
        typename = typespec[0]
        basenames = reversed(typespec)

        base = self.config.basetype
        for basename in basenames:
            defined = self._builder._get_constructor(basename, base=base)
            if isinstance(defined, type):
                base = defined

        known = {'ast': ast, 'exp': ast}
        return self._builder._instanceof(
            typename, known, ast, *args[1:], base=base, **kwargs,
        )



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import pickle

from tatsu.builder import ModelBuilderSemantics
from tatsu.tool import compile
from tatsu.util import asjson


def test_synth_model():
    grammar = """
        start::ASeq
            =
            aseq
            $
            ;

        aseq
            =
            {'a'}+
            ;
    """

    m = compile(grammar, 'ASeq')
    model = m.parse('a a a', semantics=ModelBuilderSemantics())
    assert type(model).__name__ == 'ASeq'

    p = pickle.dumps(model)
    new_model = pickle.loads(p)
    assert type(new_model).__name__ == 'ASeq'

    assert model.ast == new_model.ast


def test_nested_class_synth_model():
    grammar = """
        start::ASeq
            =
            seqs:aseq
            $
            ;

        aseq::Seq
            =
            values:{'a'}+
            ;
    """

    m = compile(grammar, 'ASeq')
    model = m.parse('a a a', semantics=ModelBuilderSemantics())
    assert type(model).__name__ == 'ASeq'

    p = pickle.dumps(model)
    new_model = pickle.loads(p)
    assert type(new_model).__name__ == 'ASeq'

    # NOTE: Since we are unpickling an object which contains nested objects, we can't do
    #       self.assertEqual(model.ast, new_model.ast) as the memory locations will be different.
    #       So either (1) we recursively walk the objects and compare fields or (2) we convert it into a
    #       str()/repr()/JSON and compare that. The latter as it is easier.
    print(f'{type(model)=}')
    assert asjson(model.ast) == asjson(new_model.ast), type(model)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import datetime
import keyword
import os
import os.path
from pathlib import Path
from typing import Any


def startscript() -> str:
    import __main__ as main

    if main.__package__:
        return main.__package__
    elif isinstance(main.__file__, str):
        return Path(main.__file__).name
    else:
        return 'unknown'


def is_posix() -> bool:
    return os.name == 'posix'


def identity(*args: Any) -> Any:
    if len(args) == 1:
        return args[0]
    return args


def format_if(fmt, values):
    return fmt % values if values else ''


def timestamp():
    return '.'.join(
        '%2.2d' % t for t in datetime.datetime.now(datetime.UTC).utctimetuple()[:-2]
    )


try:
    import psutil
except ImportError:

    def memory_use():
        return 0

else:

    def memory_use():  # pyright: ignore[reportRedeclaration]
        process = psutil.Process(os.getpid())
        return process.memory_info().rss


def try_read(filename):
    if isinstance(filename, Path):
        filename = str(filename)
    for e in ['utf-16', 'utf-8', 'latin-1', 'cp1252', 'ascii']:
        try:
            return Path(filename).read_text(encoding=e)
        except UnicodeError:
            pass
    raise ValueError(f"cannot find the encoding for '{filename}'")


def filelist_from_patterns(patterns, ignore=None, base='.', sizesort=False):
    ignore = ignore or ()
    base = Path(base or '.').expanduser()

    filenames = set()
    for pattern in patterns or []:
        path = base / pattern
        if path.is_file():
            filenames.add(path)
            continue

        if path.is_dir():
            path += '/*'

        parts = path.parts[1:] if path.is_absolute() else path.parts
        joined_pattern = str(Path().joinpath(*parts))
        filenames.update(
            p for p in Path(path.root).glob(joined_pattern) if not p.is_dir()
        )

    filenames = list(filenames)

    def excluded(path):
        if any(path.match(ex) for ex in ignore):
            return True

        return any(
            any(Path(part).match(ex) for ex in ignore or ()) for part in path.parts
        )

    if ignore:
        filenames = [path for path in filenames if not excluded(path)]
    if sizesort:
        filenames.sort(key=lambda f: f.stat().st_size)

    return filenames


def short_relative_path(path: str | Path, base: str | Path = '.') -> Path:
    path = Path(path)
    base = Path(base)
    common = Path(os.path.commonpath([base.resolve(), path.resolve()]))

    if common == path.root:
        return path
    elif common == Path.home():
        up = Path('~')
    elif common == base:
        up = Path()
    else:
        n = len(base.parts) - len(common.parts)
        up = Path('../' * n)

    rel = up / path.resolve().relative_to(common)
    if len(str(rel)) < len(str(path)):
        return rel
    else:
        return path


def fqn(obj: Any) -> str:
    # by [apalala@gmail.com](https://github.com/apalala)
    # by Gemini (2026-01-30)

    """Helper to safely retrieve the fully qualified name of a callable."""
    module = getattr(obj, "__module__", None)
    qualname = getattr(obj, "__qualname__", None)

    if module and qualname and module != "builtins":
        return f"{module}.{qualname}"
    return qualname or str(obj)


def is_reserved(name) -> bool:
    return (
        keyword.iskeyword(name)
        or keyword.issoftkeyword(name)
        or name in {'type', 'list', 'dict', 'set'}
    )


def typename(obj: Any) -> str:
    return type(obj).__name__



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import pytest

import tatsu
from tatsu.exceptions import FailedParse
from tatsu.ngcodegen import pythongen
from tatsu.util import trim

EXEC = 'exec'


def test_whitespace_directive():
    grammar = """
        @@whitespace :: /[\t ]+/

        test = "test" $;
    """
    model = tatsu.compile(grammar, 'test')
    code = pythongen(model)
    compile('test.py', code, EXEC)


def test_whitespace_none_directive():
    grammars = [
        """
            @@whitespace :: None
            @@nameguard :: False

            test = {'x'}+ $;
        """,
        """
            @@whitespace :: False
            @@nameguard :: False

            test = {'x'}+ $;
        """,
    ]
    for grammar in grammars:
        model = tatsu.compile(grammar, 'test')
        assert model.parse('xx', trace=True) == ['x', 'x']
        with pytest.raises(FailedParse):
            model.parse('x x', trace=True)


def test_whitespace_escaping():
    grammar = r'''
    @@grammar::Markdown
    @@whitespace :: /[ ]/
    start = pieces $ ;
    text = text:/[a-z]+/ ;
    pieces = {text}* ;
    '''

    with pytest.raises(FailedParse):
        tatsu.parse(grammar, '[]')


def test_default_whitespace():
    grammar = r"""
        start = {'x'}+ $;
    """

    tatsu.parse(grammar, "x x x")


def test_eol_comments_re_directive():
    grammar = """
        @@eol_comments :: /#.*?$/

        test = "test" $;
    """
    model = tatsu.compile(grammar, 'test')
    code = pythongen(model)
    compile(code, 'test.py', EXEC)


def test_left_recursion_directive():
    grammar = """
        @@left_recursion :: False

        test = "test" $;
    """
    model = tatsu.compile(grammar, 'test')
    assert not model.directives.get('left_recursion')
    assert not model.config.left_recursion

    code = pythongen(model)
    compile('test.py', code, EXEC)


def test_whitespace_no_newlines():
    grammar = """
        @@whitespace :: /[\t ]+/
        # this is just a token with any character but space and newline
        # it should finish before it capture space or newline character
        token = /[^ \n]+/;
        # expect whitespace to capture spaces between tokens, but newline
        # should be captured afterwards
        token2 = {token}* /\n/;
        # document is just list of this strings of tokens
        document = {@+:token2}* $;
    """
    text = trim("""\
        a b
        c d
        e f
    """)

    expected = [(['a', 'b'], '\n'), (['c', 'd'], '\n'), (['e', 'f'], '\n')]

    model = tatsu.compile(grammar, 'document')
    ast = model.parse(text, start='document')
    assert expected == ast


def test_grammar_directive():
    grammar = """
        @@grammar :: Test

        start = test $;
        test = "test";
    """
    model = tatsu.compile(grammar)
    assert model.directives.get('grammar') == 'Test'
    assert model.name == 'Test'

    code = pythongen(model)
    module = compile(code, 'test.py', EXEC)

    assert 'TestParser' in module.co_names


def test_parseinfo_directive():
    grammar = """
        @@parseinfo
        @@parseinfo :: True

        test = value:"test" $;
    """
    model = tatsu.compile(grammar, 'test')
    ast = model.parse('test')
    assert ast.parseinfo is not None

    code = pythongen(model)
    print(code)
    assert 'parseinfo=True' in code
    compile(code, 'test.py', EXEC)

    grammar = """
        @@parseinfo :: False

        test = value:"test" $;
    """
    model = tatsu.compile(grammar, 'test')
    ast = model.parse('test')
    assert ast.parseinfo is None

    code = pythongen(model)
    assert 'parseinfo=False' in code
    compile(code, 'test.py', EXEC)


def test_nameguard_directive():
    grammar = """
        @@grammar :: test
        @@nameguard :: False
        @@namechars :: ''

        start = sequence $ ;
        sequence = {digit}+ ;
        digit = 'x' | '1' | '2' | '3' | '4' | '5' ;
    """

    model = tatsu.compile(grammar)
    assert not model.config.nameguard
    assert model.parse('23') == ['2', '3']
    assert model.parse('xx') == ['x', 'x']



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys
import warnings
from typing import Any

from .string import prints


def stderr_print(*args: Any, **kwargs: Any) -> None:
    file = kwargs.pop('file', sys.stderr)
    print(*args, file=file, **kwargs)


def info(*args: Any, **kwargs: Any) -> None:
    stderr_print('ⓘ', *args, **kwargs)


def debug(*args: Any, **kwargs: Any) -> None:
    if not __debug__:
        return
    stderr_print('🐞', *args, **kwargs)


def warning(*args: Any, **kwargs: Any) -> None:
    warnings.warn(prints(*args, **kwargs), stacklevel=2)


def error(*args: Any, **kwargs: Any) -> None:
    raise RuntimeError(prints(*args, **kwargs))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import json  # noqa: F401
from typing import Any

import pytest

import tatsu
from tatsu import grammars
from tatsu.objectmodel import Node


def test_node_kwargs() -> None:
    class Atom(Node):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        arguments: Any = None
        symbol: Any = None

    atom = Atom(symbol='foo')
    assert type(atom).__name__ == 'Atom'
    assert atom.ast is None
    assert atom.symbol == 'foo'

    atom = Atom(symbol='foo', ast={})
    assert atom.ast == {}
    assert atom.symbol is not None
    assert atom.symbol == 'foo'

    atom = Atom(ast=42, symbol='foo')
    assert atom.ast == 42
    assert atom.symbol is not None
    assert atom.symbol == 'foo'

    atom = Atom(ast={'bar': 1}, symbol='foo')
    assert atom.ast == {'bar': 1}

    with pytest.raises(TypeError, match=r'Overriding method children'):
        atom = Atom(children=[])
        assert atom


def test_children():
    grammar = r"""
        @@grammar::Calc


        start
            =
            expression $
            ;


        expression
            =
            | add:addition
            | sub:subtraction
            | term:term
            ;


        addition::Add
            =
            left:term op:'+' ~ right:expression
            ;


        subtraction::Subtract
            =
            left:term op:'-' ~ right:expression
            ;


        term
            =
            | mul:multiplication
            | div:division
            | factor:factor
            ;


        multiplication::Multiply
            =
            left:factor op:'*' ~ right:term
            ;


        division::Divide
            =
            left:factor '/' ~ right:term
            ;


        factor
            =
            | subexpression
            | number
            ;


        subexpression
            =
            '(' ~ @:expression ')'
            ;


        number::int
            =
            /\d+/
            ;
    """

    parser = tatsu.compile(grammar, asmodel=True, trace=False)
    assert parser
    model = parser.parse('3 + 5 * ( 10 - 20 )')
    assert model
    assert model['add']
    assert model['add'].children()
    assert type(model['add'].children()[0]).__name__ == 'Multiply'


def test_model_repr():
    node = Node()
    assert repr(node) == 'Node()'

    node = Node('hello')
    assert repr(node) == "Node(ast='hello')"

    node = Node(ast='hello')
    assert repr(node) == "Node(ast='hello')"

    node = Node(ast='hello', ctx='world')
    assert repr(node) == "Node(ast='hello')"

    with pytest.raises(ValueError, match=r'world='):
        Node(ast='hello', world='world')

    token = grammars.Token(ast='hello')
    assert repr(token) == "Token(token='hello')"

    token = grammars.Token(token='hello')
    assert repr(token) == "Token(token='hello')"

    grammars.Token('hello')
    assert repr(token) == "Token(token='hello')"

    with pytest.raises(TypeError, match=r'unexpected keyword argument'):
        grammars.Token(x='x')

    with pytest.raises(TypeError, match=r'name is required'):
        grammars.Named()

    named = grammars.Named(name='foo')
    assert repr(named) == "Named(name='foo', exp=Void())"

    named = grammars.Named(name='foo', ast='bar')
    assert repr(named) == "Named(name='foo', exp='bar')"

    named = grammars.Named(ast={'name': 'foo', 'exp': 'bar'})
    assert repr(named) == "Named(name='foo', exp='bar')"

    named = grammars.Named(name='foo', exp='bar')
    assert repr(named) == "Named(name='foo', exp='bar')"



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import argparse

from .common import filelist_from_patterns
from .parproc import processing_loop


def parallel_test_run(parse, options):
    def pysearch(pattern):
        if pattern.endswith('.py'):
            return pattern
        else:
            if not pattern.endswith('/'):
                pattern += '/'
            if '**' not in pattern:
                pattern += '**/'
            return pattern + '*.py'

    try:
        patterns = [pysearch(p) for p in options.patterns]
        filenames = filelist_from_patterns(
            patterns,
            sizesort=options.sort,
            ignore=options.ignore,
        )

        kwargs = vars(options)
        kwargs.pop('patterns', None)
        kwargs.pop('sort', None)
        kwargs.pop('ignore', None)
        parallel = not kwargs.pop('serial', False)

        return processing_loop(filenames, parse, parallel=parallel, **kwargs)

    except KeyboardInterrupt:
        if options.verbose:
            raise


def parse_args():
    argparser = argparse.ArgumentParser(
        prog=__package__,
        # description=DESCRIPTION,
        add_help=False,
    )

    argparser.add_argument(
        'patterns',
        nargs='+',
        metavar='PATTERNS',
        help='filename patterns',
    )
    argparser.add_argument(
        '--help',
        '-h',
        help='show this help message and exit',
        action='help',
    )
    argparser.add_argument(
        '--ignore',
        '-i',
        metavar='PATTERN',
        help='ignore these patterns',
        action='append',
    )
    argparser.add_argument(
        '--sort',
        '-s',
        help='sort files by size',
        action='store_true',
    )
    argparser.add_argument(
        '--serial',
        '-S',
        help='do not run in parallel',
        action='store_true',
    )
    argparser.add_argument(
        '--trace',
        '-t',
        help='produce verbose output for a parse',
        action='store_true',
    )
    argparser.add_argument(
        '--verbose',
        '-v',
        help='output exceptions as they happen',
        action='store_true',
    )
    argparser.add_argument(
        '--exitfirst',
        '-x',
        help='output exceptions as they happen',
        action='store_true',
    )
    return argparser.parse_args()


def generic_main(parse):
    return parallel_test_run(parse, parse_args())



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from pathlib import Path

from tatsu import railroads
from tatsu.tool import api


def test_railroads():
    grammar = Path('./grammar/tatsu.tatsu').read_text()
    model = api.compile(grammar)
    print('RAILROADS')
    railroads.draw(model)

    tracks = railroads.tracks(model)
    assert len(tracks) == 243

    track0 = "start ●─grammar─■"
    assert tracks[0] == track0
    trackm2 = "eof∷EOF ●─'$' ✂ ──■"
    assert tracks[-2] == trackm2
    assert not tracks[-1].rstrip()



from .api import (
    compile,
    gencode,
    genmodel,
    parse,
    to_python_model,
    to_python_sourcecode,
)
from .cli import tatsu_main

__all__ = [
    'compile',
    'gencode',
    'genmodel',
    'tatsu_main',
    'parse',
    'to_python_model',
    'to_python_sourcecode',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .boundeddict import BoundedDict
from .lrudict import LRUBoundedDict

__all__: list[str] = ['BoundedDict', 'LRUBoundedDict']



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import pytest

from tatsu.objectmodel import Node, tatsudataclass


def test_init_attributes():
    node = Node()
    assert hasattr(node, 'ast')
    assert hasattr(node, 'parseinfo')


def test_init_attributes_transferred():
    node = Node(ast='Hello!')
    assert node.ast == 'Hello!'

    node = Node(ctx=object(), ast='Hello!')
    assert node.ast == 'Hello!'
    assert node.ctx


def test_attributes_through_shell():
    node = Node(ast='Hello')

    assert hasattr(node, 'ast')
    assert hasattr(node, 'parseinfo')


@tatsudataclass
class Inner(Node):
    id: str


@tatsudataclass
class Outer(Node):
    left: Inner
    right: Inner


def test_children():
    with pytest.raises(TypeError):
        Outer()

    with pytest.raises(TypeError):
        Inner()

    with pytest.raises(TypeError):
        Inner('x')

    a_inner = Inner(id='a')
    b_inner = Inner(id='b')
    outer = Outer(left=a_inner, right=b_inner)
    assert outer
    assert isinstance(outer.left, Inner)
    assert isinstance(outer.right, Inner)
    children = outer.children()
    assert set(children) == {a_inner, b_inner}



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import math
import shutil
from collections.abc import Callable
from pathlib import Path
from typing import Any

from invoke import (  # pyright: ignore[reportMissingImports, reportPrivateImportUsage]
    Context,  # pyright: ignore[reportMissingImports, reportPrivateImportUsage]
    Result,  # pyright: ignore[reportMissingImports, reportPrivateImportUsage]
    Task,  # pyright: ignore[reportMissingImports, reportPrivateImportUsage]
    task,  # pyright: ignore[reportMissingImports, reportPrivateImportUsage]
    )

__copyright__: str = 'Copyright (c) 2017-2026 Juancarlo Añez'
__license__: str = 'BSD-4-Clause'

# by Gemini 2026-02-15
# Defeat `ruff --fix` replacing `3.14` with `math.pi`
# and breaking havock on `uv run --python PYTHON`
# Fun using ord of the hidden STX char for '2'
# PYTHON: float = float(f'{math.pi:.{ord('')}f}')
PYTHON: float = 3.15

LINE_PRE: int = 4
THIN_LINE: str = '─'
THICK_LINE: str = '━'

type TaskFun = Task | Callable[[Context], Result | None] | None


def uv_python_pin(c: Context) -> float:
    result = c.run('uv python pin', pty=True, hide='both')
    if result is None:
        return PYTHON
    return float(result.stdout.strip())


def uv(c: Context, cmd: str, args: str | list[str], *, quiet: bool = True, python: float = PYTHON, group: str = 'dev',
        nogroup: str = '', **kwargs: Any, ) -> Result:
    uvpython = uv_python_pin(c)
    q = ' --quiet' if quiet else ''
    p = f' --python {python!s}' if python and python != uvpython else ''
    g = f' --group {group}' if group else ''
    n = f' --no-group {nogroup}' if nogroup else ''

    options = kwargs
    args = ' '.join(args) if isinstance(args, list) else args
    return c.run(f'uv {cmd}{q}{p}{g}{n} {args}', **options) or Result()


def uv_run(c: Context, args: str | list[str], *, python: float = PYTHON, group: str = 'dev', quiet: bool = True,
        **kwargs: Any, ) -> Result:
    return uv(c, 'run', args=args, python=python, group=group, quiet=quiet, **kwargs)


def uv_sync(c: Context):
    uv_run(c, 'sync', group='dev', quiet=True)


def version_python(c: Context, python: float = PYTHON) -> str:
    return uv_run(c, 'python3 --version', python=python, quiet=True, hide='both', ).stdout.strip()


def version_tatsu(c: Context, python: float = PYTHON) -> str:
    return uv_run(c, 'python3 -m tatsu --version', python=python, quiet=True, hide='both', ).stdout.strip()


def boundary_print(banner: str = '', line: str = THIN_LINE):
    cols = shutil.get_terminal_size().columns
    if not banner:
        print(line * cols)
    else:
        pre = LINE_PRE
        add = sum(ord(c) >= 256 for c in banner)
        print(line * pre, banner, line * (cols - 2 - pre - add - len(banner)))


def success_print(target: str = '', *, task: TaskFun = None, line: str = THIN_LINE):
    target += task.name if task else ''  # ty:ignore[unresolved-attribute]  # pyright: ignore[reportFunctionMemberAccess]
    boundary_print(f'✔ {target}', line=line)


def version_boundary_print(c: Context, target: str = '', python: float = PYTHON, line: str = THICK_LINE, ):
    verpython = version_python(c, python=python)
    vertatsu = version_tatsu(c, python=python)
    boundary_print(f'{target} {verpython} {vertatsu}', line=line)


@task
def begin(c: Context):
    boundary_print()


@task(pre=[begin])
def clean(c: Context, plus: bool = False):
    print('-> clean')
    patterns = ['build', 'dist', 'tmp', 'tatsu.egg-info', '.tox']
    if plus:
        patterns.extend(['.cache', '.pytest_cache', '.ruff_cache', '.mypy_cache'])

    for p in patterns:
        path = Path(p)
        if path.exists():
            if path.is_dir():
                shutil.rmtree(path)
            else:
                path.unlink()

    for p in Path().rglob('__pycache__'):
        shutil.rmtree(p)


@task(pre=[clean])
def ruff(c: Context, python: float = PYTHON):
    print('-> ruff')
    uv_run(c, 'ruff check -q --preview tatsu tests examples', python=python, group='test', )


@task(pre=[clean])
def ty(c: Context, python: float = PYTHON):
    print('-> ty')
    res = uv_run(c, 'ty check tatsu tests examples', python=python, group='test', warn=True, hide='both', )

    if res.exited != 0 or 'All checks passed!' not in res.stdout:
        for r in [res.stdout, res.stderr]:
            if r.strip():
                print(r)


@task(pre=[clean])
def pyright(c: Context, python: float = PYTHON):
    print('-> pyright')
    uv_run(c, 'basedpyright tatsu tests examples', python=python, group='test', pty=True, hide='both', )


@task(pre=[clean])
def pytest_fast(c: Context, python: float = PYTHON):
    print('-> pytest fast')
    Path('./tmp').mkdir(exist_ok=True)
    Path('./tmp/__init__.py').touch()
    uv_run(c, ['pytest', '--quiet', '-n auto', 'tests/', '--ignore-glob=tests/z*',
        ], python=python, group='test', hide='stdout', pty=True, )


@task(pre=[clean])
def pytest_bootstrap(c: Context, python: float = PYTHON):
    print('-> pytest bootstrap')
    Path('./tmp').mkdir(exist_ok=True)
    Path('./tmp/__init__.py').touch()
    uv_run(c, ['pytest', '--quiet', 'tests/z_bootstrap_test.py',
        ], python=python, group='test', hide='stdout', pty=True, )


@task(pre=[clean, pytest_fast, pytest_bootstrap])
def pytest(c: Context, python: float = PYTHON):
    pass


@task(pre=[clean])
def black(c: Context, python: float = PYTHON):
    print('-> black')
    res = uv_run(c, ["black", "--check", "tatsu", "tests", "examples"], python=python, group='test', warn=True,
        hide=True, pty=True, )
    if res.exited != 0:
        print(res.stdout.splitlines()[-1])
        print('✖ failed!')


@task(pre=[begin, clean, ruff, ty, pyright, black])
def lint(c: Context):
    success_print(task=lint)


@task(pre=[begin, lint, pytest])
def test(c: Context):
    success_print(task=test)


@task(pre=[clean])
def doclint(c: Context, python: float = PYTHON):
    print('-> doclint')
    uv_run(c, 'vale README.rst docs/*.rst', group='doc', hide='stdout', pty=True, )


@task(pre=[begin, doclint])
def docs(c: Context):
    print('-> docs')
    with c.cd('docs'):
        uv_run(c, 'make -s html', quiet=True, group='doc', hide='stdout')
    success_print(task=docs)


@task(pre=[clean])
def build(c: Context):
    print('-> build')
    c.run('uvx hatch build', hide='both')
    success_print(task=build)


def matrix_core(c: Context, python: float = PYTHON):
    version_boundary_print(c, target='ᝰ', python=python)
    ruff(c, python=python)
    ty(c, python=python)
    pyright(c, python=python)
    black(c, python=python)
    pytest(c, python=python)
    success_print(str(python))


@task
def py312(c: Context):
    matrix_core(c, python=3.12)
    uv_sync(c)


@task
def py313(c: Context):
    matrix_core(c, python=3.13)
    uv_sync(c)


@task
def py314(c: Context):
    matrix_core(c, python=round(math.pi, 2))
    uv_sync(c)


@task
def py315(c: Context):
    matrix_core(c, python=3.15)
    uv_sync(c)


@task(pre=[py312, py313, py314, py315])
def matrix(c: Context):
    uv_sync(c)
    success_print(task=matrix, line=THICK_LINE)


def _export_requirements(c: Context, filename: str, group: str = '', nogroup: str = ''):
    out_file = Path(filename)
    print(f'-> {out_file}')

    # note:
    #   We use pty=True here to ensure the shell redirection behaves
    #   and we see the output immediately if there's an error.
    uv(c, 'export', f'--no-hashes --format requirements-txt -o {out_file}', group=group, nogroup=nogroup, quiet=True,
        pty=True, )


@task
def req_base(c: Context):
    _export_requirements(c, 'requirements.txt', nogroup='dev')


@task
def req_dev(c: Context):
    _export_requirements(c, 'requirements-dev.txt', group='dev')


@task
def req_test(c: Context):
    _export_requirements(c, 'requirements-test.txt', group='test', nogroup='dev')


@task
def req_doc(c: Context):
    _export_requirements(c, 'requirements-doc.txt', group='doc', nogroup='dev')


@task(pre=[begin, req_base, req_dev, req_test, req_doc])
def requirements(c: Context):
    success_print(task=requirements)


@task(pre=[requirements])
def reqs(c: Context):
    pass


@task(pre=[build])
def testpublish(c: Context):
    print('-> test publish')
    workflow = 'test_publish.yml'
    c.run(f'gh workflow run {workflow}')
    c.run(f'gh run list --workflow={workflow}')


@task(pre=[build])
def publish(c: Context, dry_run: bool = True):
    print('-> publish')
    workflow = 'publish.yml'
    c.run(f'gh workflow run {workflow}')
    c.run(f'gh run list --workflow={workflow}')


@task
def g2e(c: Context):
    print('-> examples/g2e')
    with c.cd('examples/g2e'):
        c.run('uv run make -s clean test', pty=True, hide='both')
        c.run('uv run make -s clean', pty=True, hide='both')


@task
def calc(c: Context):
    print('-> examples/calc')
    with c.cd('examples/calc'):
        c.run('uv run make -s clean test', pty=True, hide='both')


@task(pre=[clean, begin, g2e, calc])
def examples(c: Context):
    success_print(task=examples)


@task(pre=[test, docs, examples, build, requirements], default=True)
def all(c: Context):
    uv_sync(c)
    boundary_print('✔ all')



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# by [apalala@gmail.com](https://github.com/apalala)
# by Gemini (2026-01-29)

from __future__ import annotations

import functools
import inspect
import warnings
from collections.abc import Callable
from typing import Any

from .common import fqn

type Decorator = Callable[[Callable[..., Any]], Callable[..., Any]]
type AnyCallable = Callable[..., Any]


def deprecated(
    replacement: AnyCallable | None = None,
) -> Callable[[AnyCallable], AnyCallable]:
    # by [apalala@gmail.com](https://github.com/apalala)
    # by Gemini (2026-01-30)

    def decorator(func: AnyCallable) -> AnyCallable:
        # Get FQN for the deprecated function and its replacement
        func_name = fqn(func)

        if replacement is not None:
            repl_name = fqn(replacement)
            msg = f"\nDEP `{func_name}()` is deprecated! \nUSE `{repl_name}()` instead."
        else:
            msg = f"\n`DEP {func_name}()` is deprecated and will be removed in a future version."

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            warnings.warn(message=msg, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def deprecated_params(**params_map: str | None) -> Decorator:
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not params_map:
                return func(*args, **kwargs)

            sig = inspect.signature(func)
            bound_args = sig.bind_partial(*args, **kwargs)
            actual_args = bound_args.arguments

            # Access __name__ safely to satisfy the linter
            func_name = getattr(func, "__name__", str(func))

            for old_p, new_p in params_map.items():
                if old_p not in actual_args:
                    continue

                msg = f"Parameter '{old_p}' is deprecated."
                if new_p:
                    msg += f" Use '{new_p}' instead."
                warnings.warn(
                    message=f"{msg} (in {func_name})",
                    category=DeprecationWarning,
                    stacklevel=2,
                )

                if not new_p:
                    continue

                if new_p not in actual_args:
                    kwargs[new_p] = actual_args.pop(old_p)
                    continue

                actual_args.pop(old_p)
                warnings.warn(
                    message=f"Both '{old_p}' and '{new_p}' provided. Using '{new_p}'.",
                    category=RuntimeWarning,
                    stacklevel=2,
                )

            return func(*args, **kwargs)

        return wrapper

    return decorator



#!/usr/bin/env python3
# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# by [apalala@gmail.com](https://github.com/apalala)
# by Gemini (2026-02-03)
from __future__ import annotations

import subprocess
import sys
from collections.abc import Collection
from datetime import date
from pathlib import Path

from common import get_staged_files


def is_header_missing(path: Path, target: Collection[str]) -> bool:
    """
    Check if a file is missing the license header using native string ops.
    Works for # (Python/Makefile) and // (JSONC) comments.
    """
    try:
        with path.open('r', encoding='utf-8', errors='ignore') as f:
            # Check the first 1024 bytes for the header
            head = f.read(1024)
            return any(line not in head for line in target)
    except Exception:
        return False


def main() -> None:
    """
    Main entry point. Exits with 1 if any staged files lack the header.
    """
    year = date.today().year
    target = {
        'Copyright (c)',
        f'{year} Juancarlo Añez (apalala@gmail.com)',
        'SPDX-License-Identifier: BSD-4-Clause',
    }
    ignored_suffix = {
        '.dot',
        '.g4',
        '.ico',
        '.jpg',
        '.lock',
        '.md',
        '.pdf',
        '.png',
        '.pyc',
        '.txt',
        '.zip',
    }

    ignored_prefix = [
        'bootstrap',
    ]

    ignored_paths = [
        Path('./.vale/styles/'),
    ]

    staged = get_staged_files()
    missing_paths: list[Path] = []

    for path in staged:
        must_ignore = (
            path.suffix in ignored_suffix
            or any(path.stem.startswith(p) for p in ignored_prefix)
            or any(path.is_relative_to(p) for p in ignored_paths)
        )
        if must_ignore:
            continue

        if is_header_missing(path, target):
            missing_paths.append(path)

    if missing_paths:
        print(
            "ERROR: Commit aborted. The following files are missing the license header:"
        )
        for f in missing_paths:
            print(f"  - {f}")
        print(f"\nPlease add:\n{target}")
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations

from tatsu.tool import parse


def test_constant_interpolation():
    input = '42 69'
    grammar = r"""
            start = a:number b: number i:`"seen: {a}, {b}"` $ ;
            number = /\d+/ ;
    """
    assert parse(grammar, input) == {'a': '42', 'b': '69', 'i': 'seen: 42, 69'}


def test_constant_interpolation_free():
    input = '42 69'
    grammar = r"""
            start = a:number b: number i:`seen: {a}, {b}` $ ;
            number = /\d+/ ;
    """
    assert parse(grammar, input) == {'a': '42', 'b': '69', 'i': 'seen: 42, 69'}


def test_constant_interpolation_multiline():
    input = '42 69'
    grammar = r"""
            start = a:number b: number
            i:```
            seen:
            {a}
            {b}
            ``` $ ;
            number = /\d+/ ;
    """

    result = parse(grammar, input)
    print(result)
    assert result == {'a': '42', 'b': '69', 'i': 'seen:\n42\n69'}


def test_evaluate_constant():
    grammar = """
        @@grammar :: constants
        start = int bool str null 'a' $;

        int = `42` ;
        bool = `True` ;
        str = `Something` ;
        null = `None` ;
    """

    ast = parse(grammar, 'a')
    assert ast == (42, True, 'Something', None, 'a')



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from tatsu.exceptions import GrammarError
from tatsu.tool import compile
from tatsu.util.asjson import asjsons


def test_missing_rule():
    grammar = """
        @@grammar::TestGrammar
          block = test ;
    """
    try:
        model = compile(grammar)
        print('MODEL', asjsons(model))
        model.parse('abc')
    except GrammarError as e:
        assert str(e) == 'Unknown rules, no parser generated:\ntest'


def test_missing_rules():
    grammar = """
        @@grammar::TestGrammar
          block = test | test2;
    """
    try:
        model = compile(grammar)
        print('MODEL', asjsons(model))
        model.parse('abc')
    except GrammarError as e:
        assert str(e) == 'Unknown rules, no parser generated:\ntest\ntest2'



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import unittest

from tatsu.exceptions import FailedSemantics
from tatsu.grammars import ModelContext
from tatsu.tool import compile


class StatefulTests(unittest.TestCase):
    def test_stateful(self):
        # Parser for mediawiki-style unordered lists.
        grammar = r"""
            document = @:ul [ nl ] $ ;
            ul = "*" ul_start el+:li { nl el:li } * ul_end ;
            li = ul | li_text ;
            (* Quirk: If a text line is followed by a sublist, the sublist does not get its own li. *)
            li_text = text:text [ ul:li_followed_by_ul ] ;
            li_followed_by_ul = nl @:ul ;
            text = /.*/ ;
            nl = /\n/ ul_marker ;
            (* The following rules are placeholders for state transitions. *)
            ul_start = () ;
            ul_end = () ;
            (* The following rules are placeholders for state validations and grammar rules. *)
            ul_marker = () ;
            """

        class StatefulSemantics:
            def __init__(self, parser):
                self._context = parser

            def ul_start(self, ast):
                ctx = self._context
                ctx.substate = 1 if ctx.substate is None else ctx.substate + 1
                return ast

            def ul_end(self, ast):
                ctx = self._context
                ctx.substate = (
                    None
                    if ctx.substate is None or ctx.substate <= 1
                    else ctx.substate - 1
                )
                return ast

            def ul_marker(self, ast):
                ctx = self._context
                if ctx.substate is not None and not ctx.tokenizer.match(
                    '*' * ctx.substate,
                ):
                    raise FailedSemantics('not at correct level')
                return ast

            def ul(self, ast):
                return '<ul>' + ''.join(ast.el) + '</ul>'

            def li(self, ast):
                return '<li>' + ast + '</li>'

            def li_text(self, ast):
                return ast.text if ast.ul is None else ast.text + ast.ul

        model = compile(grammar, 'test')
        context = ModelContext(model.rules, whitespace='', nameguard=False)

        ast = model.parse(
            '*abc',
            start='document',
            ctx=context,
            semantics=StatefulSemantics(context),
            whitespace='',
            nameguard=False,
        )
        self.assertEqual(ast, '<ul><li>abc</li></ul>')

        ast = model.parse(
            '*abc\n',
            start='document',
            ctx=context,
            semantics=StatefulSemantics(context),
            whitespace='',
            nameguard=False,
        )
        self.assertEqual('<ul><li>abc</li></ul>', ast)

        ast = model.parse(
            '*abc\n*def\n',
            start='document',
            ctx=context,
            semantics=StatefulSemantics(context),
            whitespace='',
            nameguard=False,
        )
        self.assertEqual('<ul><li>abc</li><li>def</li></ul>', ast)

        ast = model.parse(
            '**abc',
            start='document',
            ctx=context,
            semantics=StatefulSemantics(context),
            whitespace='',
            nameguard=False,
        )
        self.assertEqual('<ul><li><ul><li>abc</li></ul></li></ul>', ast)

        ast = model.parse(
            '*abc\n**def\n',
            start='document',
            ctx=context,
            semantics=StatefulSemantics(context),
            whitespace='',
            nameguard=False,
        )
        self.assertEqual('<ul><li>abc<ul><li>def</li></ul></li></ul>', ast)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause

# NOTE: re-exports mostly for bw compatibility
from .abctools import *  # noqa: F403
from .asjson import *  # noqa: F403
from .common import *  # noqa: F403
from .debugging import *  # noqa: F403
from .misc import *  # noqa: F403
from .string import *  # noqa: F403
from .typing import *  # noqa: F403
from .undefined import *  # noqa: F403



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import unittest

from tatsu.exceptions import FailedParse
from tatsu.ngcodegen import pythongen
from tatsu.tool import compile
from tatsu.util import trim


class PatternTests(unittest.TestCase):
    def test_patterns_with_newlines(self):
        grammar = """
            @@whitespace :: /[ \t]/
            start
                =
                blanklines $
                ;

            blanklines
                =
                blankline [blanklines]
                ;

            blankline
                =
                /(?m)^[^\\n]*\\n$/
                ;
        """

        model = compile(grammar, 'test')
        ast = model.parse('\n\n')
        self.assertEqual(('\n', '\n'), ast)

    def test_pattern_concatenation(self):
        grammar = """
            start
                =
                {letters_digits}+
                ;


            letters_digits
                =
                ?"[a-z]+"
                + ?'[0-9]+'
                ;
        """
        pretty = """
            start
                =
                {letters_digits}+
                ;


            letters_digits
                =
                /[a-z]+/
                + /[0-9]+/
                ;
        """
        model = compile(grammar=grammar)
        ast = model.parse('abc123 def456')
        self.assertEqual(['abc123', 'def456'], ast)
        print(model.pretty())
        self.assertEqual(trim(pretty), model.pretty())

    def test_ignorecase_not_for_pattern(self):
        grammar = """
            @@ignorecase
            start
                =
                {word} $
                ;

            word
                =
                /[a-z]+/
                ;
        """

        model = compile(grammar=grammar)
        try:
            model.parse('ABcD xYZ')
            self.fail('@@ignorecase should not apply to patterns')
        except FailedParse:
            pass

    def test_ignorecase_pattern(self):
        grammar = """
            start
                =
                {word} $
                ;

            word
                =
                /(?i)[a-z]+/
                ;
        """

        model = compile(grammar=grammar)
        ast = model.parse('ABcD xYZ')
        self.assertEqual(['ABcD', 'xYZ'], ast)

    def test_multiline_pattern(self):
        grammar = r"""
            start =
            /(?x)
            foo
            bar
            / $ ;
        """
        model = compile(grammar=trim(grammar))
        from tatsu import grammars

        assert isinstance(model.rules[0].exp, grammars.Sequence)
        print(pythongen(model.rules[0].exp.sequence[0]))
        self.assertEqual(
            pythongen(model.rules[0].exp.sequence[0]).strip(),
            "self._pattern(r'(?x)\\nfoo\\nbar\\n')",
        )

        grammar = r"""
            start =
            /(?x)foo\nbar
            blort/ $ ;
        """
        model = compile(grammar=trim(grammar))
        assert isinstance(model.rules[0].exp, grammars.Sequence)
        print(pythongen(model.rules[0].exp.sequence[0]))
        self.assertEqual(
            trim(pythongen(model.rules[0].exp.sequence[0])),
            "self._pattern(r'(?x)foo\\nbar\\nblort')",
        )



#!/usr/bin/env python3

# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause

"""
Log parser to split, reverse, and rejoin git-style logs.
Diagnostic messages are sent to stderr; reversed text is sent to stdout.
by [apalala@gmail.com](https://github.com/apalala)
Gemini 3 Flash - 2026-01-28
"""
from __future__ import annotations

import os
import re
import sys


def parse_commits(log_text):
    # by [apalala@gmail.com](https://github.com/apalala)
    pattern = re.compile(r'^> commit', re.MULTILINE)
    starts = [m.start() for m in pattern.finditer(log_text)]
    ends = starts[1:] + [len(log_text)]

    return [log_text[start:end].strip() for start, end in zip(starts, ends)]


def main():
    # by [apalala@gmail.com](https://github.com/apalala)
    if len(sys.argv) < 2:
        print(f"Usage: {sys.argv[0]} <log_file>", file=sys.stderr)
        sys.exit(1)

    filename = sys.argv[1]

    if not os.path.isfile(filename):
        print(f"Error: File '{filename}' not found.", file=sys.stderr)
        sys.exit(1)

    try:
        with open(filename, 'r', encoding='utf-8') as f:
            log_content = f.read()
    except Exception as e:
        print(f"Error reading file: {e}", file=sys.stderr)
        sys.exit(1)

    commits = parse_commits(log_content)

    # Send diagnostic info to stderr
    print(f"Parsed {len(commits)} commits from {filename}.", file=sys.stderr)

    # Reverse the list of commits
    commits.reverse()

    # Rejoin the content with spacing for readability
    output_text = "\n\n".join(commits)

    # Print the actual text output to stdout
    print(output_text)


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Iterable
from contextlib import contextmanager
from copy import copy
from typing import Any, Self

from ..ast import AST
from ..infos import Alert

__all__ = ['ParseState', 'ParseStateStack']

from ..util.abctools import is_list


class ParseState:
    __slots__ = ('alerts', 'ast', 'cst', 'last_node', 'pos')

    def __init__(self, pos: int = 0, ast: Any = None, cst: Any = None):
        self.pos: int = pos
        self.ast: Any = ast or AST()
        self.cst: Any = cst
        self.last_node: Any = None
        self.alerts: list[Alert] = []

    @property
    def node(self) -> Any:
        ast = self.ast
        cst = self.cst
        if not ast:
            return tuple(cst) if is_list(cst) else cst
        elif '@' in ast:
            return ast['@']
        else:
            return ast


class ParseStateStack:
    def __init__(self: Self) -> None:
        self._state_stack: list[ParseState] = [ParseState()]
        self._cut_stack: list[bool] = [False]

    @property
    def top(self) -> ParseState:
        return self._state_stack[-1]

    @property
    def state(self) -> ParseState:
        return self.top

    @property
    def ast(self) -> Any:
        return self.top.ast

    @ast.setter
    def ast(self, value: Any) -> None:
        self.top.ast = value

    @property
    def cst(self) -> Any:
        return self.top.cst

    @cst.setter
    def cst(self, value: Any) -> Any:
        self.top.cst = value
        return self.cst

    @property
    def pos(self) -> int:
        return self.top.pos

    @property
    def last_node(self) -> Any:
        return self.top.last_node

    @last_node.setter
    def last_node(self, value: Any) -> None:
        self.top.last_node = value

    @property
    def node(self) -> Any:
        return self.top.node

    def pop(self) -> ParseState:
        return self._state_stack.pop()

    def push(self, pos: int, ast: Any = None, cst: Any = None) -> ParseState:
        ast = copy(self.ast) if ast is None else ast
        self.state.pos = pos

        newstate = ParseState(pos=pos, ast=ast, cst=cst)
        self._state_stack.append(newstate)

        return self.top

    def merge(self, pos: int) -> ParseState:
        prev = self.pop()
        self.ast = prev.ast
        self.extend(prev.cst)
        return prev

    def alert(self, level: int = 1, message: str = '') -> Alert:
        self.top.alerts.append(Alert(level=level, message=message))
        return self.top.alerts[-1]

    @staticmethod
    def copy_node(node: Any) -> Any:
        if node is None:
            return None
        elif is_list(node):
            return node[:]
        else:
            return node

    def append(self, node: Any) -> Any:
        self.last_node = node
        previous = self.cst
        if previous is None:
            self.cst = self.copy_node(node)
        elif is_list(previous):
            previous.append(node)
        else:
            self.cst = [previous, node]
        return node

    def extend(self, node: Any) -> Any:
        self.last_node = node
        if node is None:
            return None
        previous = self.cst
        if previous is None:
            self.cst = self.copy_node(node)
        elif is_list(node):
            if is_list(previous):
                previous.extend(node)
            else:
                self.cst = [previous, *node]
        elif is_list(previous):
            previous.append(node)
        else:
            self.cst = [previous, node]
        return node

    def setname(self, name: str) -> None:
        self.ast._set(name, self.last_node)

    def addname(self, name: str) -> None:
        self.ast._setlist(name, self.last_node)

    def define(
        self,
        keys: Iterable[str],
        list_keys: Iterable[str] | None = None,
    ) -> Any:
        ast = AST()
        ast._define(keys, list_keys=list_keys)
        ast.update(self.ast)
        self.ast = ast
        return self.ast

    def cut_seen(self) -> bool:
        return self._cut_stack[-1]

    def push_cut(self):
        self._cut_stack.append(False)

    def pop_cut(self) -> bool:
        return self._cut_stack.pop()

    def set_cut_seen(self, prune: bool = True) -> None:
        self._cut_stack[-1] = True

    @contextmanager
    def cutscope(self):
        self.push_cut()
        try:
            yield
        finally:
            self.pop_cut()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Iterable, Iterator

from ..grammars import Grammar, Rule

type RuleName = str
type Headmap = dict[RuleName, set[RuleName]]
type SCC = set[RuleName]  # Strongly Connected Component


class GrammarAnalysis:
    def __init__(self, grammar: Grammar) -> None:
        self.grammar = grammar
        self.head_map = self._build_head_map()
        self.grammar_order = {r.name: i for i, r in enumerate(self.grammar.rules)}
        self.sccs = self._build_sccs()
        self.recursive_rules = self._find_recursive_rules()

    def _get_rule(self, name: str) -> Rule:
        return self.grammar.rulemap[name]

    def rule_in_order_key(self, name: str) -> int:
        return self.grammar_order[name]

    def ordered_rule_names(self, names: Iterable[RuleName]) -> list[RuleName]:
        return sorted(names, key=self.rule_in_order_key)

    def _build_head_map(self) -> Headmap:
        def is_rule_name(element: str) -> bool:
            return element in self.grammar.rulemap

        return {
            rule.name: {
                str(la[0]) for la in rule.lookahead() if la and is_rule_name(str(la[0]))
            }
            for rule in self.grammar.rules
        }

    def _find_recursive_rules(self) -> set[RuleName]:
        def in_own_headmap(name) -> bool:
            return name in self.head_map.get(name, set())

        return set(filter(in_own_headmap, (r.name for r in self.grammar.rules)))

    # NOTE: Tarjan's Algorithm for Strongly Connected Components
    def _build_sccs(self) -> list[SCC]:
        index = 0
        stack: list[RuleName] = []
        indices: dict[RuleName, int] = {}
        lowlink: dict[RuleName, int] = {}
        on_stack: set[RuleName] = set()
        sccs: list[SCC] = []

        # Define what counts as a "Rule" vs a "Terminal"
        valid_rules = [r.name for r in self.grammar.rules]

        def strongconnect(v: RuleName) -> None:
            nonlocal index
            indices[v] = index
            lowlink[v] = index
            index += 1
            stack.append(v)
            on_stack.add(v)

            for w in self.head_map.get(v, set()):
                # Only traverse if the dependency is also a defined rule
                if w in valid_rules:
                    if w not in indices:
                        strongconnect(w)
                        lowlink[v] = min(lowlink[v], lowlink[w])
                    elif w in on_stack:
                        lowlink[v] = min(lowlink[v], indices[w])

            if lowlink[v] == indices[v]:
                component = set()
                while True:
                    w = stack.pop()
                    on_stack.remove(w)
                    component.add(w)
                    if w == v:
                        break
                sccs.append(component)

        # Only start the walk from actual rules
        for rule_name in valid_rules:
            if rule_name not in indices:
                strongconnect(rule_name)

        return sccs

    # CAVEAT: this does not work in all leftrec cases
    def find_leaders_in_scc(self, scc: SCC) -> set[RuleName]:
        leaders = set(scc)
        for start in scc:
            # Try to find a leader such that all cycles go through it.
            for cycle in self.find_cycles_in_scc(scc, start):
                if len(cycle) > 2:
                    # leaders -= set(scc) - set(cycle)
                    leaders &= set(cycle)
                # if not leaders:
                #     break
                # raise ValueError(
                #     f"SCC {scc} {leaders} has no leadership candidate"
                #     " (no element is included in all cycles)"
                # )
        return leaders

    def find_cycles_in_scc(self, scc: SCC, start: RuleName) -> list[list[RuleName]]:
        """Find cycles in SCC emanating from start.

        Yields lists of the form ['A', 'B', 'C', 'A'], which means there's
        a path from A -> B -> C -> A.  The first item is always the start
        argument, but the last item may be another element, e.g.  ['A',
        'B', 'C', 'B'] means there's a path from A to B and there's a
        cycle from B to C and back.
        """
        # Basic input checks.
        assert start in scc, (start, scc)
        assert start in self.head_map, (start, scc)

        # Recursive helper that yields cycles.
        def dfs(node: RuleName, path: list[RuleName]) -> Iterator[list[str]]:
            if node in path:  # found a cycle
                yield [*path, node]
                return
            path = [*path, node]
            for child in self.head_map[node]:
                yield from dfs(child, path)

        return list(dfs(start, []))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause

from __future__ import annotations

import builtins
from collections import namedtuple

from .. import grammars
from ..mixins.indent import IndentPrintMixin
from ..objectmodel import Node
from ..util import safe_name
from ..util.abctools import compress_seq
from ..util.deprecate import deprecated_params
from ..util.itertools import topsort

HEADER = """\
    #!/usr/bin/env python3

    # CAVEAT UTILITOR
    #
    # This file was automatically generated by 竜TatSu.
    #
    #    https://pypi.python.org/pypi/tatsu/
    #
    # Any changes you make to it will be overwritten the next time
    # the file is generated.

    # ruff: noqa: PLC2401, PLC2402, PLC2403
    # fmt: off

    from __future__ import annotations

    from dataclasses import dataclass
    from typing import Any

    from tatsu.objectmodel import tatsudataclass
    from tatsu.builder import ModelBuilderSemantics, types_defined_in
    {basetype_import}


    class {name}ModelBuilderSemantics(ModelBuilderSemantics):
        def __init__(self, constructors=None, **kwargs):
            constructors = constructors or []
            constructors += types_defined_in(globals())
            super().__init__(basetype={basetype}, constructors=constructors, **kwargs)


    class {name}ModelBuilderSemantics({name}ModelBuilderSemantics):
        ...
"""


BaseClassSpec = namedtuple('BaseClassSpec', ['class_name', 'base'])


@deprecated_params(base_type='basetype')
def modelgen(
    model: grammars.Grammar,
    name: str = '',
    basetype: type = Node,
    base_type: type | None = None,
) -> str:
    if isinstance(base_type, type):
        basetype = base_type

    generator = PythonModelGenerator(name=name, basetype=basetype)
    return generator.generate_model(model)


class PythonModelGenerator(IndentPrintMixin):

    def __init__(
        self,
        name: str = '',
        basetype: type = Node,
        base_type: type | None = None,
    ):
        if isinstance(base_type, type):
            basetype = base_type
        basetype = basetype or Node
        super().__init__()
        self.basetype = basetype
        self.name = name or None

    def generate_model(self, grammar: grammars.Grammar):
        basetype = self.basetype
        basetype_name = basetype.__name__.split('.')[-1]
        basetype_import = f"from {basetype.__module__} import {basetype_name}"

        self.name = self.name or grammar.name
        self.print(
            HEADER.format(
                name=self.name,
                basetype=self.basetype.__name__,
                basetype_import=basetype_import,
            ),
        )

        rule_index = grammar.rulemap
        rule_specs = {rule.name: self._base_class_specs(rule) for rule in grammar.rules}
        rule_specs = {name: specs for name, specs in rule_specs.items() if specs}

        specs_by_name = {
            s.class_name: s.base for specs in rule_specs.values() for s in specs
        }
        base = self._model_base_name()
        specs_by_name[base] = basetype_name

        all_specs = {
            (s.class_name, s.base) for specs in rule_specs.values() for s in specs
        }
        model_names = list(reversed(topsort(specs_by_name, all_specs)))

        model_to_rule = {
            rule_specs[name][0].class_name: rule
            for name, rule in rule_index.items()
            if name in rule_specs
        }

        for model_name in model_names:
            if model_name in vars(builtins):
                continue
            if rule := model_to_rule.get(model_name):
                self._gen_rule_class(rule, rule_specs[rule.name])
            else:
                self._gen_base_class(model_name, specs_by_name.get(model_name))

        return self.printed_text()

    @staticmethod
    def _model_base_name():
        return 'ModelBase'

    def _print_dataclass(self):
        self.print('@tatsudataclass')

    def _gen_base_class(self, class_name: str, base: str | None):
        self.print()
        self.print()
        self._print_dataclass()
        if base:
            self.print(f'class {class_name}({base}):')
        else:
            self.print(f'class {class_name}:')
        with self.indent():
            self.print('pass')

    def _gen_rule_class(self, rule: grammars.Rule, specs: list[BaseClassSpec]):
        if not specs:
            return
        spec = specs[0]
        arguments = sorted({safe_name(d) for d, _ in compress_seq(rule.defines())})

        self.print()
        self.print()
        self._print_dataclass()
        self.print(f'class {spec.class_name}({spec.base}):')
        with self.indent():
            if not arguments:
                self.print('pass')
            for arg in arguments:
                self.print(f'{arg}: Any = None')

    def _base_class_specs(self, rule: grammars.Rule) -> list[BaseClassSpec]:
        if not rule.params or not isinstance(rule.params[0], str):
            return []
        spec = rule.params[0].split('::')
        base = [self._model_base_name()]
        class_names = [safe_name(n) for n in spec] + base
        return [
            BaseClassSpec(class_name, class_names[i + 1])
            for i, class_name in enumerate(class_names[:-1])
        ]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from ..util import unicode_display_len as ulen

type Rails = list[str]

ETX = '␃'


def assert_one_length(rails: Rails) -> Rails:
    if not rails:
        return []

    length = ulen(rails[0])

    def one_length(rail: str) -> bool:
        check = isinstance(rail, str) and ulen(rail) == length
        assert check, f'ulen({rail=!r}) != {length=}'
        return check

    assert all(one_length(rail) for rail in rails), 'lengths differ'
    return rails


def pad(rrl: str, c: str, maxl: int) -> str:
    return f'{rrl}{c * (maxl - ulen(rrl))}'


def railpad_(rrl: str, maxl: int) -> str:  # name_ for visual alignment
    return pad(rrl, '─', maxl=maxl)


def blankpad(rrl: str, maxl: int) -> str:
    return pad(rrl, ' ', maxl=maxl)


def lay_out(tracks: list[Rails]) -> Rails:
    # by Gemini 2026/02/17

    if not tracks:
        return []
    if len(tracks) == 1:
        return tracks[0]

    maxl = max(ulen(p[0]) if p else 0 for p in tracks)
    out: Rails = []

    for rails in tracks[:-1]:
        assert isinstance(rails, list), f'{rails=!r}'
        if not rails:
            continue

        joint = rails[0]
        if ETX not in joint:
            out += [f"  ├─{railpad_(joint, maxl)}─┤ "]
        else:
            out += [f"  ├─{blankpad(joint, maxl)} │ "]

        for rail in rails[1:]:
            out += [f"  │ {blankpad(rail, maxl)} │ "]

    # the last set of rails
    rails = tracks[-1]
    joint = rails[0]
    if ETX not in joint:
        out += [f"  └─{railpad_(joint, maxl)}─┘ "]
    else:
        corner = "─┘ "
        out[-1] = out[-1][: -len(corner)] + corner
        out += [f"  └─{blankpad(joint, maxl)}   "]

    for rail in rails[1:]:
        out += [f"    {blankpad(rail, maxl)}   "]

    # the first rail
    joint = tracks[0][0] if tracks[0] else ''
    if ETX not in joint:
        out[0] = f"──┬─{railpad_(joint, maxl)}─┬─"
    else:
        out[0] = f"──┬─{blankpad(joint, maxl)} ┬─"

    return assert_one_length(out)


def looptail(rails: Rails, maxl: int) -> Rails:
    out = []
    for line in rails:
        out += [f"  │ {blankpad(line, maxl)} │  "]

    out += [f"  └─{railpad_('', maxl)}<┘  "]

    return assert_one_length(out)


def stopnloop(rails: Rails) -> Rails:
    # by Gemini 2026/02/17
    if not rails:
        return ["───>───"]

    maxl = max(ulen(line) for line in rails)
    out = []

    first = rails[0]
    out += [f"──┬─{railpad_(first, maxl)}─┬──"]

    out += looptail(rails[1:], maxl)
    return assert_one_length(out)


def loop(rails: Rails) -> Rails:
    # by Gemini 2026/02/17
    if not rails:
        return ["───>───"]

    maxl = max(ulen(line) for line in rails)
    out = []

    out += [f"──┬→{railpad_('', maxl)}─┬──"]

    first = rails[0]
    out += [f"  ├→{railpad_(first, maxl)}─┤  "]  # xxx

    out += looptail(rails[1:], maxl)
    return assert_one_length(out)


def weldtwo(left: Rails, right: Rails) -> Rails:
    assert isinstance(left, list), f'{left=!r}'
    assert isinstance(right, list), f'{right=!r}'

    if not right or ETX in left:
        return left[:]
    if not left:
        return right[:]

    len_left = ulen(left[0])
    len_right = ulen(right[0])
    out_height = max(len(left), len(right))
    common_height = min(len(left), len(right))

    out = left[:]
    for i in range(out_height):
        if i < common_height:
            out[i] += right[i]
        elif i < len(out):
            out[i] += f'{'':{len_right}}'
        else:
            out += [f'{'':{len_left}}{right[i]}']

    return assert_one_length(out)


def weld(*tracks: Rails) -> Rails:
    if not tracks:
        return []
    out = tracks[0][:]
    for rails in tracks[1:]:
        out = weldtwo(out, rails)
    return assert_one_length(out)



import unittest
import warnings


class MiscTests(unittest.TestCase):
    # test that Mapping is imported from collections.abc
    # required in python 3.8.0+
    def test_import_mapping_from_collectionsABC(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            self.assertEqual(len(w), 0)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import builtins
import inspect
import itertools
from collections.abc import Callable, Iterable, Mapping, Sequence
from dataclasses import dataclass, field, replace
from types import ModuleType
from typing import Any

from .itertools import CycleError, first, topsort

__all__ = [
    'ActualArguments',
    'BoundCallable',
    'Constructor',
    'TypeContainer',
    'boundcall',
]

type Constructor = type | Callable
type TypeContainer = type | ModuleType | Mapping[str, type] | dict[str, type]


def boundcall(fun: Callable, known: dict[str, Any], *args: Any, **kwargs: Any) -> Any:
    return BoundCallable.call(fun, known, *args, **kwargs)


@dataclass
class ActualArguments:
    """
    A DTO representing the resolved positional and keyword arguments for a call.
    """

    # by [apalala@gmail.com](https://github.com/apalala)
    # by Gemini (2026-02-09)

    args: list[Any] = field(default_factory=list)
    arg_names: dict[str, Any] = field(default_factory=dict)
    kwargs: dict[str, Any] = field(default_factory=dict)

    def seen(self, name: str) -> bool:
        return name in self.arg_names or name in self.kwargs

    def empty(self) -> bool:
        return not (bool(self.args) or bool(self.kwargs))

    def add_arg(self, name: str, var: Any):
        if name in self.arg_names:
            return
        self.arg_names[name] = var
        self.args.append(var)

    def add_kwarg(self, name: str, var: Any):
        if name in self.arg_names:
            return
        self.kwargs[name] = var

    def add_args(self, args: Iterable[Any]):
        self.args.extend(args)

    def add_kwargs(self, kwargs: Mapping[str, Any]):
        self.kwargs.update(kwargs)

    def unique(self) -> ActualArguments:
        clone = replace(self)
        for name in list(clone.kwargs):
            if name in clone.arg_names:
                del clone.kwargs[name]
        return clone


class BoundCallable:
    # by [apalala@gmail.com](https://github.com/apalala)
    # by Gemini (2026-02-09)
    """
    Handles the binding of context-aware data to a callable's signature.
    """

    def __init__(self, fun: Callable, known: dict[str, Any], *args: Any, **kwargs: Any):
        self.fun = fun
        self.actual = self.bind(fun, known, *args, **kwargs)

    def __call__(self) -> Any:
        return self.fun(*self.actual.args, **self.actual.kwargs)

    @staticmethod
    def call(fun: Callable, known: dict[str, Any], *args: Any, **kwargs: Any) -> Any:
        actual = BoundCallable.bind(fun, known, *args, **kwargs)
        return fun(*actual.args, **actual.kwargs)

    @staticmethod
    def bind(
        fun: Callable,
        known: dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> ActualArguments:
        arg = next(iter(known.values())) if known else (args[0] if args else None)

        funname = getattr(fun, '__name__', None)
        if funname in vars(builtins):
            return ActualArguments(args=[arg])

        declared = inspect.signature(fun).parameters
        p = inspect.Parameter

        actual = ActualArguments()
        argsc = list(args)
        kwargsc = kwargs.copy()
        for name, param in declared.items():
            if actual.seen(name):
                continue

            match param.kind:
                case p.KEYWORD_ONLY:
                    actual.add_kwarg(name, kwargsc.pop(name, None))
                case p.POSITIONAL_OR_KEYWORD:
                    if name in kwargsc:
                        actual.add_kwarg(name, kwargsc.pop(name, None))
                    else:
                        actual.add_arg(name, argsc.pop(0) if argsc else arg)
                case p.POSITIONAL_ONLY:
                    actual.add_arg(
                        name,
                        argsc.pop(0) if argsc else arg,
                    )  # note: inject known arg
                case p.VAR_POSITIONAL:
                    actual.add_args(argsc)
                case p.VAR_KEYWORD:
                    actual.add_kwargs(kwargs)

        for name, value in known.items():
            if not (param := declared.get(name, None)):
                continue
            match param.kind:
                case p.KEYWORD_ONLY:
                    actual.add_kwarg(name, value)
                case p.POSITIONAL_OR_KEYWORD:
                    actual.add_kwarg(name, value)
                case p.POSITIONAL_ONLY:
                    actual.add_arg(name, value)

        return actual.unique()


def least_upper_bound_type(constructors: Sequence[Constructor]) -> type:
    # by [apalala@gmail.com](https://github.com/apalala)
    # by Gemini (2026-01-30)

    # Caller is responsible for filtering constructors to relevant types
    types_ = [t for t in constructors if isinstance(t, type)]

    if not types_:
        return object
    if len(types_) == 1:
        return types_[0]

    nodes: set[type] = set()
    edges: set[tuple[type, type]] = set()

    for t in types_:
        # mro[1:] focuses on the skeleton/ancestors
        ancestors = t.__mro__[1:]
        nodes.update(ancestors)

        edges.update((child, parent) for child, parent in itertools.pairwise(ancestors))

    if not nodes:
        return object

    try:
        topsorted = topsort(list(nodes), list(edges))
    except CycleError:
        return object

    # The LUB is the most specific ancestor that covers all provided types
    # Since child -> parent, we check from the most specific in the sort
    return first(
        (parent for parent in topsorted if all(issubclass(t, parent) for t in types_)),
        default=object,
    )



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any, cast

__all__ = ['UndefinedType', 'Undefined', 'notnone']


def notnone[T](value: T | None, default: T) -> T:
    return value if value is not None else default


class UndefinedType[T]:
    __notnone: Any = None

    def __init__(self):
        if isinstance(self.__notnone, UndefinedType):
            return
        type(self).__notnone = self

    @classmethod
    def notnone(cls) -> UndefinedType[T]:
        return cast(UndefinedType[T], cls.__notnone)

    def __eq__(self, other: Any) -> bool:
        if self is not Undefined:
            return False
        if other is not Undefined:
            return False
        return other is self

    def __bool__(self):
        return False

    def __and__(self, other):
        return self

    def __rand__(self, other):
        return self.__and__(other)

    def __or__(self, other):
        return other

    def __ror__(self, other):
        return self.__or__(other)

    def __invert__(self):
        return not None

    def __repr__(self) -> str:
        return 'Undefined'

    def __str__(self) -> str:
        return 'Undefined'

    def __hash__(self) -> int:
        return hash(id(self))


Undefined = UndefinedType()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import subprocess
from pathlib import Path


def get_staged_files() -> list[Path]:
    """
    Get the list of files currently staged for commit.
    """
    try:
        result = subprocess.run(
            ['git', 'diff', '--cached', '--name-only', '--diff-filter=d'],
            capture_output=True,
            text=True,
            check=True,
        )
        return [Path(filename) for filename in result.stdout.splitlines()]
    except subprocess.CalledProcessError:
        return []



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
Unit tests for safe_eval using the 'check' naming convention.
Generated by Gemini on 2026-01-25.
"""

import ast

import pytest

from tatsu.util.safeeval import SecurityError, check_safe_eval, safe_eval


class MockObj:
    def __init__(self, val):
        self.val = val

    def do_it(self):
        return self.val


def test_simple_context_eval():
    context = {'a': 1, 'b': 2}
    expression = 'seen: {a} {b}'
    expected = "seen: 1 2"

    with pytest.raises(SecurityError, match='Invalid expression syntax'):
        safe_eval(expression, context)
        check_safe_eval(expression, context)
        result = safe_eval(expression, context)
        assert result == expected

    with pytest.raises(SyntaxError):
        result = ast.literal_eval(expression)  # Should not raise
        assert result == expected

    eval_expression = f'{"f" + repr(expression.strip())}'
    result = safe_eval(eval_expression, context)
    assert result == expected


def test_safe_eval_flow():
    ctx = {"MockObj": MockObj, "x": 10}
    assert safe_eval("MockObj(x).do_it()", ctx) == 10


def test_unauthorized_call():
    with pytest.raises(SecurityError, match="Unauthorized function call"):
        check_safe_eval("open('file.txt')", {"x": 1})


def test_dunder_block():
    with pytest.raises(SecurityError, match="Dunder access prohibited"):
        check_safe_eval("x.__class__", {"x": 1})


def test_name_mismatch():
    with pytest.raises(SecurityError, match="Context name mismatch"):
        check_safe_eval("fake_fn()", {"fake_fn": abs})  # abs.__name__ is 'abs'


def test_no_excetions_for_eval():
    class MyException(BaseException):
        pass

    context = {"MyException": MyException}
    with pytest.raises(SecurityError, match="Exception class forbidden at context"):
        check_safe_eval("1 + 1", context)
        safe_eval("1 + 1", context)


def test_bad_argcount():
    context = {'object': object, 'type': type}
    with pytest.raises(SecurityError, match="Bad argument count"):
        safe_eval("type('Evil', (object,), {})", context)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
import subprocess  # noqa: S404


def test_cli_help():
    output = subprocess.check_output(['tatsu', '--help'])  # noqa: S607
    output = output.decode('utf-8')
    pattern = r'(?ms)竜TatSu takes a grammar .*GRAMMAR'
    assert re.search(pattern, output)


def test_cli_python():
    output = subprocess.check_output(['tatsu', './grammar/tatsu.tatsu'])  # noqa: S607
    output = output.decode('utf-8')
    pattern = (
        r'(?ms)CAVEAT UTILITOR.*?竜TatSu.*?KEYWORDS: set\['
        r'.*?class \w*?Parser\(Parser\):'
    )
    assert re.search(pattern, output)


def test_cli_model():
    output = subprocess.check_output(
        ['tatsu', '-g', './grammar/tatsu.tatsu'],  # noqa: S607
    )
    output = output.decode('utf-8')
    pattern = (
        r'(?ms)CAVEAT UTILITOR.*?竜TatSu'
        r'.*?class \w+?ModelBuilderSemantics\(ModelBuilderSemantics\):'
        r'.*?class \w+?ModelBuilderSemantics\(\w+?ModelBuilderSemantics\):'
    )
    assert re.search(pattern, output)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any

from ..exceptions import FailedLeftRecursion
from ..infos import RuleInfo
from ..parserconfig import ParserConfig
from ..tokenizing import Tokenizer
from ..util import color, info
from ..util.unicode_characters import C_CUT, C_ENTRY, C_FAILURE, C_RECURSION, C_SUCCESS


class EventColor(color.Color):
    def __init__(self):
        self.ENTRY = self.YELLOW + self.BRIGHT + C_ENTRY
        self.SUCCESS = self.GREEN + self.BRIGHT + C_SUCCESS
        self.FAILURE = self.RED + self.BRIGHT + C_FAILURE
        self.RECURSION = self.RED + self.BRIGHT + C_RECURSION
        self.CUT = self.MAGENTA + self.BRIGHT + C_CUT


C = EventColor()


class EventTracer:
    def trace(self, msg: str, *args: Any, **kwargs: Any) -> None: ...

    def trace_event(self, event: str) -> None: ...

    def trace_entry(self) -> None: ...

    def trace_success(self) -> None: ...

    def trace_failure(self, ex: Exception | None = None) -> None: ...

    def trace_recursion(self) -> None: ...

    def trace_cut(self) -> None: ...

    def trace_match(
        self,
        token: Any,
        name: str | None = None,
        failed: bool = False,
    ) -> None: ...


class EventTracerImpl(EventTracer):
    def __init__(
        self,
        tokenizer: Tokenizer,
        ruleinfos: list[RuleInfo],
        *,
        config: ParserConfig | None = None,
        **settings: Any,
    ) -> None:
        self.config = ParserConfig.new(config, **settings)
        self.tokenizer = tokenizer

        # NOTE: not copying / sharing same list with caller
        self.ruleinfos = ruleinfos

        if self.config.colorize:
            color.init()
        global C  # noqa: PLW0603
        C = EventColor()

    def trace(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if not self.config.trace:
            return

        info(msg, *args, **kwargs)

    def trace_event(self, event: str) -> None:
        if not self.config.trace:
            return

        fname = ''
        if self.config.trace_filename:
            fname = self.tokenizer.lineinfo().filename
        if fname:
            fname += '\n'

        lookahead = self.tokenizer.lookahead().rstrip()
        lookahead = '\n' + lookahead if lookahead else ''

        message = (
            f'{event}{self.rulestack()}'
            f' {C.DIM}{fname}'
            f'{self.tokenizer.lookahead_pos()}{C.RESET}'
            f'{C.RESET_ALL}{lookahead}{C.RESET_ALL}'
        )
        self.trace(message)

    def trace_entry(self) -> None:
        self.trace_event(f'{C.ENTRY}')

    def trace_success(self) -> None:
        self.trace_event(f'{C.SUCCESS}')

    def trace_failure(self, ex: Exception | None = None) -> None:
        if isinstance(ex, FailedLeftRecursion):
            self.trace_recursion()
        else:
            self.trace_event(f'{C.FAILURE}')

    def trace_recursion(self) -> None:
        self.trace_event(f'{C.RECURSION}')

    def trace_cut(self) -> None:
        self.trace_event(f'{C.CUT}')

    def trace_match(
        self,
        token: Any,
        name: str | None = None,
        failed: bool = False,
    ) -> None:
        if not self.config.trace:
            return

        name_str = f'/{name}/' if name else ''
        if self.config.trace_filename:
            fname = self.tokenizer.lineinfo().filename + '\n'
        else:
            fname = ''

        mark = f'{C.FAILURE}' if failed else f'{C.SUCCESS}'

        lookahead = self.tokenizer.lookahead().rstrip()
        lookahead = '\n' + lookahead if lookahead else lookahead

        message = (
            f'{mark}'
            f"'{token}{name_str}"
            f'{C.DIM}{fname}'
            f'{C.RESET_ALL}{lookahead}{C.RESET_ALL}'
        )
        self.trace(message)

    def rulestack(self) -> str:
        rulestack = [r.name for r in reversed(self.ruleinfos)]
        stack = self.config.trace_separator.join(rulestack)
        if (
            max((len(s) for s in stack.splitlines()), default=0)
            > self.config.trace_length
        ):
            stack = stack[: self.config.trace_length]
            stack = stack.rsplit(self.config.trace_separator, 1)[0]
            stack += self.config.trace_separator
        return stack



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .walker import draw, text, tracks

__all__ = ['draw', 'text', 'tracks']



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys
from importlib import resources
from pathlib import Path

from .. import compile, grammars
from .semantics import ANTLRSemantics


def antlr_grammar() -> str:
    path = resources.files('tatsu.g2e') / 'antlr.tatsu'
    return path.read_text()


def translate(
    text: str | None = None,
    filename: str | None = None,
    name: str | None = None,
    encoding: str = 'utf-8',
    trace: bool = False,
) -> grammars.Grammar:
    if text is None and filename is None:
        raise ValueError('either `text` or `filename` must be provided')
    if filename:
        filepath = Path(filename)

    if text is None:
        name = name or filepath.stem
        text = filepath.read_text(encoding=encoding)

    name = name or 'Unknown'

    semantics = ANTLRSemantics(name)
    grammar = compile(antlr_grammar())
    return grammar.parse(
        text,
        name=name,
        filename=filename,
        semantics=semantics,
        trace=trace,
        colorize=True,
    )


def main():
    if len(sys.argv) < 2:
        thisprog = Path(sys.argv[0]).name
        print(thisprog)
        print('Usage:')
        print('\t', thisprog, 'FILENAME.g [--trace]')
        sys.exit(1)
    model = translate(
        filename=sys.argv[1],
        trace='--trace' in sys.argv or '-t' in sys.argv,
    )
    print(model)


if __name__ == '__main__':
    main()



import unittest

from tatsu.tool import compile


class LookaheadTests(unittest.TestCase):
    def test_skip_to(self, trace=False):
        grammar = """
            start = 'x' ab $ ;

            ab
                =
                | 'a' 'b'
                | ->'a' 'b'
                ;
        """
        m = compile(grammar, trace=trace)
        ast = m.parse('x xx yyy a b')
        self.assertEqual(('x', ('a', 'b')), ast)

        grammar = """
            start = 'x' ab $ ;

            ab
                =
                | 'a' 'b'
                | ->&'a' 'a' 'b'
                ;
        """
        m = compile(grammar, trace=trace)
        ast = m.parse('x xx yyy a b')
        self.assertEqual(('x', ('a', 'b')), ast)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any, NamedTuple, Protocol

from ..infos import RuleInfo


class MemoKey(NamedTuple):
    pos: int
    ruleinfo: RuleInfo
    state: Any


class RuleResult(NamedTuple):
    node: Any
    newpos: int
    newstate: Any


class RuleLike(Protocol):
    is_leftrec: bool = False
    is_memoizable: bool = False
    is_name: bool = False

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        pass


class closure(list[Any]):
    def __hash__(self) -> int:  # pyright: ignore[reportIncompatibleVariableOverride]
        return hash(tuple(self))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .basenode import BaseNode, TatSuDataclassParams, tatsudataclass
from .node import Node

__all__ = ['BaseNode', 'Node', 'TatSuDataclassParams', 'tatsudataclass']



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from datetime import UTC, datetime
from pathlib import Path


def iso_timestamp() -> str:
    """
    Returns an ISO 8601 string with UTC timezone (Z).
    Example: '2026-02-15T22-15-01Z'
    """
    now = datetime.now(UTC)
    isostr = now.isoformat(timespec='seconds')
    return isostr.replace(':', '-').replace('+00-00', 'Z')


def iso_logpath(prefix: str, basdir: str = "log", suffix='.log') -> Path:
    """Helper to generate a pathlib.Path using the ISO timestamp."""
    ts = iso_timestamp()
    logdir = Path(basdir)
    logdir.mkdir(parents=True, exist_ok=True)
    suffix = suffix if suffix.startswith('.') else '.' + suffix
    return logdir / f"{prefix}_{ts}{suffix}"



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import functools
import weakref
from collections.abc import Iterable, Mapping
from typing import Any

from .basenode import BaseNode, tatsudataclass

__all__ = ['Node', 'tatsudataclass']

from ..util.deprecate import deprecated


@tatsudataclass
class Node(BaseNode):
    def __init__(self, ast: Any = None, **kwargs: Any):
        super().__init__(ast=ast, **kwargs)
        self.__parent_ref: weakref.ref[Node] | None = (  # pyright: ignore[reportRedeclaration]
            None
        )

    def __post_init__(self):
        super().__post_init__()
        self.__parent_ref: weakref.ref[Node] | None = None

    @property
    def parent(self) -> Node | None:
        ref = self.__parent_ref
        if ref is None:
            return None
        else:
            return ref()

    @property
    def comments(self) -> Any:
        deprecated(replacement=None)(self.comments)
        return None

    @property
    def text(self) -> str | None:
        pi = self.parseinfo
        if pi and hasattr(pi.tokenizer, "text"):
            return pi.tokenizer.text[pi.pos : pi.endpos]
        return None

    @property
    def line(self) -> int | None:
        return self.parseinfo.line if self.parseinfo else None

    @property
    def path(self) -> tuple[Node, ...]:
        ancestors: list[Node] = [self]
        parent = self.parent
        while parent is not None:
            ancestors.append(parent)
            parent = parent.parent
        return tuple(reversed(ancestors))

    def children(self) -> tuple[Node, ...]:
        return self._cached_children

    def children_list(self) -> list[Node]:
        return list(self._cached_children)

    @functools.cached_property
    def _cached_children(self) -> tuple[Node, ...]:
        def dfs(obj: Any) -> Iterable[Node]:
            match obj:
                case Node() as node:
                    node.__parent_ref = weakref.ref(self)
                    yield node
                case Mapping() as map:
                    for name, value in map.items():
                        if name.startswith("_"):
                            continue
                        if value is None:
                            continue
                        yield from dfs(value)
                case (list() | tuple()) as seq:
                    for item in seq:
                        yield from dfs(item)
                case _:
                    pass

        return tuple(dfs(self.__pub__()))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import pytest

from tatsu.util.string import regexp


def test_patterns_quotes():
    assert regexp("'") == 'r"\'"'
    assert regexp('\'') == 'r"\'"'

    # Input: a'b -> Result: r'a\'b'
    assert regexp("a'b") == 'r"a\'b"'

    assert regexp('"') == "r'\"'"
    assert regexp('"') == r"r'" + '"' + "'"


def test_backslash_edge_cases() -> None:
    assert regexp(r"\'") == 'r"\\\'"'

    # raw string r"r'\\\''" represents exactly three backslashes
    assert regexp(r"\\'") == 'r"\\\\\'"'


def test_patterns_newlines():
    assert regexp("\n") == r"r'\n'"
    assert regexp(r"\n") == r"r'\n'"
    assert regexp(r"\\n") == r"r'\\n'"


def test_patterns_expr():
    assert regexp('[abc]') == r"r'[abc]'"

    with pytest.raises(ValueError, match=r"Invalid regex passed to regexp\(\)"):
        assert regexp('\\') == r"r'\\'"

    assert regexp(r'\\') == r"r'\\'"


def test_patterns_real():
    e = r"(REM\s|')[^\r\n]*(\r?\n|\r)"
    # Result preserves the original \r \n and escapes the '
    assert regexp(e) == 'r"(REM\\s|\')[^\\r\\n]*(\\r?\\n|\\r)"'


def test_roundtrip_verification() -> None:
    # Standard case
    assert eval(regexp("it's")) == "it's"  # noqa: S307

    # Trailing backslash case: Now safe to eval()
    with pytest.raises(ValueError, match=r"Invalid regex passed to regexp\(\)"):
        assert eval(regexp("\\")) == "\\\\"  # noqa: S307


def test_edge_cases() -> None:
    assert regexp("'''") == 'r"\'\'\'"'
    assert regexp("") == "r''"
    assert regexp(123) == "r'123'"


def test_regexp_is_runnable():
    pattern = r"\'"
    generated_code = regexp(pattern)

    # This ensures the generated code is actually valid Python
    # and evaluates back to the original pattern.
    assert eval(generated_code) == pattern  # noqa: S307



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import contextlib
import pathlib
import sys
import tempfile
import unittest

from tatsu.ngcodegen import pythongen
from tatsu.parser import TatSuParserGenerator
from tatsu.tool import compile
from tatsu.util import trim


class ParameterTests(unittest.TestCase):
    def test_keyword_params(self):
        grammar = """
            start(k1=1, k2=2)
                =
                {'a'} $
                ;
        """
        g = TatSuParserGenerator('Keywords')
        model = g.parse(grammar)
        code = pythongen(model)
        self.assertEqual('#!/usr/bin/env python3', code.splitlines()[0])

    def test_35_only_keyword_params(self):
        grammar = """
            rule(kwdA=A, kwdB=B)
                =
                'a'
                ;
        """
        model = compile(grammar, 'test')
        self.assertEqual(trim(grammar), str(model))

    def test_36_params_and_keyword_params(self):
        grammar = """
            rule(A, kwdB=B)
                =
                'a'
                ;
        """
        model = compile(grammar, 'test')
        self.assertEqual(trim(grammar), str(model))

    def test_36_param_combinations(self):
        def assert_equal(target, value):
            assert target == value

        class TC36Semantics:
            """Check all rule parameters for expected types and values"""

            def rule_positional(self, ast, p1, p2, p3, p4):
                assert_equal('ABC', p1)
                assert_equal(123, p2)
                assert_equal('=', p3)
                assert_equal('+', p4)
                return ast

            def rule_keyword(self, ast, k1, k2, k3, k4):
                assert_equal('ABC', k1)
                assert_equal(123, k2)
                assert_equal('=', k3)
                assert_equal('+', k4)
                return ast

            def rule_all(self, ast, p1, p2, p3, p4, k1, k2, k3, k4):
                assert_equal('DEF', p1)
                assert_equal(456, p2)
                assert_equal('=', p3)
                assert_equal('+', p4)
                assert_equal('HIJ', k1)
                assert_equal(789, k2)
                assert_equal('=', k3)
                assert_equal('+', k4)
                return ast

        grammar = """
            @@ignorecase::False
            @@nameguard

            start
                = {rule_positional | rule_keywords | rule_all} $ ;
            rule_positional('ABC', 123, '=', '+')
                = 'a' ;
            rule_keywords(k1=ABC, k3='=', k4='+', k2=123)
                = 'b' ;
            rule_all('DEF', 456, '=', '+', k1=HIJ, k3='=', k4='+', k2=789)
                = 'c' ;
        """

        pretty = """
            @@ignorecase :: False
            @@nameguard :: True

            start
                =
                {rule_positional | rule_keywords | rule_all} $
                ;


            rule_positional(ABC, 123, '=', '+')
                =
                'a'
                ;


            rule_keywords(k1=ABC, k3='=', k4='+', k2=123)
                =
                'b'
                ;


            rule_all(DEF, 456, '=', '+', k1=HIJ, k3='=', k4='+', k2=789)
                =
                'c'
                ;
        """

        model = compile(grammar, 'RuleArguments')
        self.assertEqual(trim(pretty), str(model))
        model = compile(pretty, 'RuleArguments')

        ast = model.parse('a b c')
        self.assertEqual(['a', 'b', 'c'], ast)
        semantics = TC36Semantics()
        ast = model.parse('a b c', semantics=semantics)
        self.assertEqual(['a', 'b', 'c'], ast)
        pythongen(model)

    def test_36_unichars(self):
        grammar = """
            start = { rule_positional | rule_keywords | rule_all }* $ ;

            rule_positional("ÄÖÜäöüß") = 'a' ;

            rule_keywords(k1='äöüÄÖÜß') = 'b' ;

            rule_all('ßÄÖÜäöü', k1="ßäöüÄÖÜ") = 'c' ;
        """

        def _trydelete(pypath, pymodule):
            module_with_path = pypath / pymodule

            with contextlib.suppress(OSError):
                module_with_path.with_suffix('.py').unlink()
            with contextlib.suppress(OSError):
                module_with_path.with_suffix('.pyc').unlink()
            with contextlib.suppress(OSError):
                module_with_path.with_suffix('.pyo').unlink()

        def assert_equal(target, value):
            self.assertEqual(target, value)

        class UnicharsSemantics:
            """Check all rule parameters for expected types and values"""

            def rule_positional(self, ast, p1):
                assert_equal('ÄÖÜäöüß', p1)
                return ast

            def rule_keyword(self, ast, k1):
                assert_equal('äöüÄÖÜß', k1)
                return ast

            def rule_all(self, ast, p1, k1):
                assert_equal('ßÄÖÜäöü', p1)
                assert_equal('ßäöüÄÖÜ', k1)
                return ast

        m = compile(grammar, 'UnicodeRuleArguments')
        ast = m.parse('a b c')
        self.assertEqual(['a', 'b', 'c'], ast)

        semantics = UnicharsSemantics()
        ast = m.parse('a b c', semantics=semantics)
        self.assertEqual(['a', 'b', 'c'], ast)

        code = pythongen(m)
        module_name = 'tc36unicharstest'
        temp_dir = pathlib.Path(tempfile.mkdtemp()) / module_name
        temp_dir.mkdir(exist_ok=True)
        py_file_path = temp_dir / f'{module_name}.py'

        pathlib.Path(py_file_path).write_text(code)
        try:
            sys.path.append(str(temp_dir))
            import tc36unicharstest  # pyright: ignore[reportMissingImports]

            assert tc36unicharstest
            _trydelete(temp_dir, module_name)
        except Exception as e:
            self.fail(e)

    def test_numbers_and_unicode(self):
        grammar = """
            rúle(1, -23, 4.56, 7.89e-11, Añez)
                =
                'a'
                ;


            rúlé::Añez
                =
                'ñ'
                ;
        """

        model = compile(grammar, 'test')
        self.assertEqual(trim(grammar), str(model))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .decorator import isname, leftrec, nomemo, rule
from .engine import ParseContext

__all__ = ['ParseContext', 'isname', 'leftrec', 'nomemo', 'rule']



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
The Buffer class provides the functionality required by a parser-driven lexer.

Line analysis and caching are done so the parser can freely move with goto(p)
to any position in the parsed text, and still recover accurate information
about source lines and content.
"""

from __future__ import annotations

import re
from pathlib import Path
from typing import Any

from .infos import ParserConfig, PosLine
from .tokenizing import LineIndexInfo, LineInfo, Tokenizer
from .util.itertools import str_from_match
from .util.misc import cached_re_compile
from .util.undefined import Undefined

DEFAULT_WHITESPACE_RE = re.compile(r'(?m)\s+')

# for backwards compatibility with existing parsers
LineIndexEntry = LineIndexInfo


class Buffer(Tokenizer):

    def __init__(
        self,
        text: str,
        *,
        config: ParserConfig | None = None,
        **settings: Any,
    ):
        config = ParserConfig.new(config=config, **settings)
        self.config = config

        text = str(text)
        self._text = self.original_text = text

        self.whitespace_re = self.build_whitespace_re(config.whitespace)
        self.nameguard = (
            config.nameguard
            if config.nameguard is not None
            else bool(self.whitespace_re) or bool(config.namechars)
        )
        self._namechar_set = set(config.namechars or '')

        self._pos = 0
        self._len = 0
        self._linecount = 0
        self._lines: list[str] = []
        self._line_index: list[LineIndexInfo] = []
        self._line_cache: list[PosLine] = []

        self._preprocess()
        self._postprocess()

    @property
    def text(self) -> str:
        return self._text

    @property
    def filename(self) -> str:
        return str(self.config.filename or '')

    @property
    def ignorecase(self) -> bool:
        return bool(self.config.ignorecase)

    @property
    def whitespace(self) -> str | None:
        return self.config.whitespace

    @staticmethod
    def build_whitespace_re(whitespace: Any) -> re.Pattern | None:
        if whitespace is Undefined:
            return DEFAULT_WHITESPACE_RE
        if whitespace in {None, ''}:
            return None
        elif isinstance(whitespace, re.Pattern):
            return whitespace
        elif whitespace:
            return cached_re_compile(whitespace)
        else:
            return None

    def _preprocess(self, /, *args: Any, **kwargs: Any):
        lines, index = self._preprocess_block(self.filename, self.text)
        self._lines = lines
        self._line_index = index
        self._text = self.join_block_lines(lines)

    def _postprocess(self):
        cache, count = PosLine.build_line_cache(self._lines, len(self.text))
        self._line_cache = cache
        self._linecount = count
        self._len = len(self.text)

    def _preprocess_block(
        self,
        name: str,
        block,
        /,
        **kwargs,
    ) -> tuple[list[str], list[LineIndexInfo]]:
        lines = self.split_block_lines(block)
        index = LineIndexInfo.block_index(name, len(lines))
        return self.process_block(name, lines, index, **kwargs)

    def split_block_lines(self, block: str) -> list[str]:
        return block.splitlines(True)

    def join_block_lines(self, lines: list[str]):
        return ''.join(lines)

    def process_block(
        self,
        name: str,
        lines: list[str],
        index: list[LineIndexInfo],
        /,
        **kwargs,
    ) -> tuple[list[str], list[LineIndexInfo]]:
        return lines, index

    def include(
        self,
        lines: list[str],
        index: list[LineIndexInfo],
        i: int,
        j: int,
        name: str,
        block: str,
        /,
        **kwargs,
    ) -> int:
        blines, bindex = self._preprocess_block(name, block, **kwargs)
        assert len(blines) == len(bindex)
        lines[i:j] = blines
        index[i:j] = bindex
        assert len(lines) == len(index)
        return j + len(blines) - 1

    def include_file(
        self,
        source: str,
        name: str,
        lines: list[str],
        index: list[LineIndexInfo],
        i: int,
        j: int,
    ) -> int:
        text, filename = self.get_include(source, name)
        return self.include(lines, index, i, j, filename, text)

    def get_include(self, source: str, filename: str) -> tuple[str, str]:
        source_path = Path(source).resolve()
        base = source_path.parent
        include = base / filename
        try:
            return include.read_text(), str(include)
        except OSError as e:
            raise ValueError(f'include not found: {include}') from e

    def replace_lines(self, i: int, j: int, name: str, block: str) -> tuple[int, str]:
        lines = self.split_block_lines(self.text)
        index = self._line_index

        endline = self.include(lines, index, i, j, name, block)

        self._text = self.join_block_lines(lines)
        self._line_index = index
        self._postprocess()

        newtext = self.join_block_lines(lines[j + 1 : endline + 2])
        return endline, newtext

    @property
    def pos(self) -> int:
        return self._pos

    @pos.setter
    def pos(self, p: int):
        self.goto(p)

    @property
    def line(self) -> int:
        return self.posline()

    @property
    def col(self) -> int:
        return self.poscol()

    def posline(self, pos: int | None = None) -> int:
        if pos is None:
            pos = self._pos
        return self._line_cache[pos].line

    def poscol(self, pos: int | None = None) -> int:
        if pos is None:
            pos = self._pos
        start = self._line_cache[pos].start
        return pos - start

    def atend(self) -> bool:
        return self._pos >= self._len

    def ateol(self) -> bool:
        return self.atend() or self.current in {'\r', '\n', None}

    @property
    def current(self) -> str | None:
        if self._pos >= self._len:
            return None
        return self.text[self._pos]

    def at(self, p: int) -> str | None:
        if p >= self._len:
            return None
        return self.text[p]

    def peek(self, n: int = 1) -> str | None:
        return self.at(self._pos + n)

    def next(self) -> str | None:
        if self.atend():
            return None
        c = self.text[self.pos]
        self.move(1)
        return c

    def goto(self, pos: int):
        self._pos = max(0, min(len(self.text), pos))

    def move(self, n: int):
        self.goto(self.pos + n)

    def _eat_regex(self, regex: str | re.Pattern) -> None:
        if not regex:
            return
        while self._matchre_fast(regex):
            pass

    def _eat_regex_list(self, regex: str | re.Pattern | None) -> list[str]:
        if not regex:
            return []

        r = cached_re_compile(regex)
        if r is None:
            return []

        def takewhile_repeat_regex():
            while x := self.matchre(r):
                yield x

        return list(takewhile_repeat_regex())

    def eat_whitespace(self) -> None:
        if self.whitespace_re:
            self._eat_regex(self.whitespace_re)

    def eat_comments(self) -> list[str]:
        return self._eat_regex_list(self.config.comments)

    def eat_eol_comments(self) -> list[str]:
        return self._eat_regex_list(self.config.eol_comments)

    def next_token(self) -> None:
        p = None
        while self._pos != p:
            p = self._pos
            self.eat_eol_comments()
            self.eat_comments()
            self.eat_whitespace()

    def skip_to(self, c: str) -> int:
        p = self._pos
        le = self._len
        while p < le and self.text[p] != c:
            p += 1
        self.goto(p)
        return self.pos

    def skip_past(self, c: str) -> int:
        self.skip_to(c)
        self.next()
        return self.pos

    def skip_to_eol(self) -> int:
        return self.skip_to('\n')

    def scan_space(self) -> bool:
        return bool(self.whitespace_re) and bool(self._scanre(self.whitespace_re))

    def is_space(self) -> bool:
        return self.scan_space()

    def is_name_char(self, c: str | None) -> bool:
        return c is not None and (c.isalnum() or c in self._namechar_set)

    def is_name(self, s: str) -> bool:
        if not s:
            return False
        goodstart = s[0].isalpha() or s[0] in self._namechar_set
        return goodstart and all(self.is_name_char(c) for c in s[1:])

    def match(self, token: str) -> str | None:
        if token is None:
            return None

        p = self.pos
        text = self.text[p : p + len(token)]
        if self.ignorecase:
            is_match = text.lower() == token.lower()
        else:
            is_match = text == token
        if not is_match:
            return None

        self.move(len(token))
        partial_match = (
            self.nameguard and self.is_name_char(self.current) and self.is_name(token)
        )
        if partial_match:
            self.goto(p)
            return None

        return token

    def _matchre_fast(self, pattern: str | re.Pattern | None):
        if not (match := self._scanre(pattern)):
            return

        self.move(len(match.group()))

    def matchre(self, pattern: str | re.Pattern) -> str | None:
        if not (match := self._scanre(pattern)):
            return None

        matched = match.group()
        token = str_from_match(match)
        self.move(len(matched))
        return token

    def _scanre(self, pattern: str | re.Pattern | None) -> re.Match[Any] | None:
        cre = cached_re_compile(pattern)
        if cre is None:
            return None
        else:
            return cre.match(self.text, self.pos)

    @property
    def linecount(self) -> int:
        return self._linecount

    def lineinfo(self, pos: int | None = None) -> LineInfo:
        if pos is None:
            pos = self._pos
        if not self._line_cache or not self._line_index:
            return LineInfo(
                filename=self.filename,
                line=0,
                col=0,
                start=0,
                end=len(self.text),
                text=self.text,
            )

        # -2 to skip over sentinel
        pos = min(pos, len(self._line_cache) - 2)
        start, line, length = self._line_cache[pos]
        end = start + length
        col = pos - start

        text = self.text[start:end]

        # only required to support includes
        n = min(len(self._line_index) - 1, line)
        filename, line = self._line_index[n]

        return LineInfo(filename, line, col, start, end, text)

    def lookahead_pos(self) -> str:
        if self.atend():
            return ''
        info = self.lineinfo()
        return '@%d:%d' % (info.line + 1, info.col + 1)

    def lookahead(self) -> str:
        if self.atend():
            return ''
        info = self.lineinfo()
        text = info.text[info.col : info.col + 1 + 80]
        return self.split_block_lines(text)[0].rstrip()

    def get_line(self, n: int | None = None) -> str:
        if n is None:
            n = self.line
        return self._lines[n]

    def get_lines(self, start: int | None = None, end: int | None = None) -> list[str]:
        if start is None:
            start = 0
        if end is None:
            end = len(self._lines)
        return self._lines[start : end + 1]

    def line_index(self, start: int = 0, end: int | None = None) -> list[LineIndexInfo]:
        if end is None:
            end = len(self._line_index)
        return self._line_index[start : 1 + end]

    def __repr__(self) -> str:
        return f'{type(self).__name__}()'

    def __json__(self, seen: set[int] | None = None) -> str | None:
        return None



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import itertools
import textwrap
from collections.abc import Iterator
from typing import Any

from .. import grammars
from ..exceptions import CodegenError
from ..mixins.indent import IndentPrintMixin
from ..objectmodel import Node
from ..util import regexp, safe_name
from ..util.abctools import compress_seq
from ..util.undefined import Undefined
from ..walkers import NodeWalker

HEADER = """\
    #!/usr/bin/env python3

    # CAVEAT UTILITOR
    #
    #  This file was automatically generated by 竜TatSu.
    #
    #     https://pypi.python.org/pypi/tatsu/
    #
    #  Any changes you make to it will be overwritten the next time
    #  the file is generated.

    # ruff: noqa: RUF100, C405, COM812, I001, F401, PLR1702, PLC2801, SIM117
    # ruff: noqa: PL2401, PLC2402, PLC2403
    # fmt: off

    from __future__ import annotations

    from tatsu.buffering import Buffer
    from tatsu.infos import ParserConfig
    from tatsu.parsing import Parser, leftrec, nomemo, isname, generic_main, rule

    __all__ = ['{parser_name}Buffer', '{parser_name}Parser', 'main']
"""

FOOTER = """\
def main(filename, **kwargs):
    if not filename or filename == '-':
        import sys

        text = sys.stdin.read()
    else:
        import pathlib

        text = pathlib.Path(filename).read_text()

    parser = {name}Parser()
    return parser.parse(text, filename=filename, **kwargs)


if __name__ == '__main__':
    import json
    from tatsu.util import asjson

    ast = generic_main(main, {name}Parser, name='{name}')
    data = asjson(ast)
    print(json.dumps(data, indent=2))
"""


def pythongen(model: Node, parser_name: str = '') -> str:
    generator = PythonParserGenerator(parser_name=parser_name)
    generator.walk(model)
    return generator.printed_text()


class PythonParserGenerator(IndentPrintMixin, NodeWalker):
    _counter: Iterator[int] = itertools.count()

    def __init__(self, parser_name: str = ''):
        super().__init__()
        self.parser_name = parser_name

    @classmethod
    def _next_n(cls) -> int:
        return next(cls._counter)

    @classmethod
    def _reset_counter(cls):
        cls._counter = itertools.count()

    def walk_default(self, node: Any) -> Any:
        return node

    def walk_Grammar(self, grammar: grammars.Grammar):
        parser_name = self.parser_name or grammar.name
        self.print(HEADER.format(parser_name=parser_name))
        self.print()
        self.print()

        self._gen_keywords(grammar)
        self._gen_buffering(grammar, parser_name)
        self._gen_parsing(grammar, parser_name)

        self.print(FOOTER.format(name=parser_name))

    def walk_Rule(self, rule: grammars.Rule):
        def param_repr(p):
            if isinstance(p, int | float):
                return str(p)
            else:
                return repr(p.split('::')[0])

        self._reset_counter()
        params = kwparams = ''
        if rule.params:
            params = ', '.join(param_repr(self.walk(p)) for p in rule.params)
        if rule.kwparams:
            kwparams = ', '.join(
                f'{k}={param_repr(self.walk(v))}' for k, v in rule.kwparams.items()
            )

        if params and kwparams:
            params = params + ', ' + kwparams
        elif kwparams:
            params = kwparams

        leftrec = '\n@leftrec' if rule.is_leftrec else ''
        nomemo = '\n@nomemo' if not rule.is_memoizable and not leftrec else ''
        isname = '\n@isname' if rule.is_name else ''

        self.print(f"""
                @rule({params})\
                {leftrec}\
                {nomemo}\
                {isname}\
                \ndef _{rule.name}_(self):
            """)
        with self.indent():
            self.print(self.walk(rule.exp))

    def walk_BasedRule(self, rule: grammars.BasedRule):
        self.walk_Rule(rule)

    def walk_Call(self, call: grammars.Call):
        self.print(f'self._{call.name}_()')

    def walk_RuleInclude(self, include: grammars.RuleInclude):
        self.walk(include.rule.exp)

    def walk_Void(self, void: grammars.Void):
        self.print('self._void()')

    def walk_Dot(self, _dot: grammars.Dot):
        self.print('self._dot()')

    def walk_Fail(self, fail: grammars.Fail):
        self.print('self._fail()')

    def walk_Cut(self, cut: grammars.Cut):
        self.print('self._cut()')

    def walk_Comment(self, comment: grammars.Comment):
        lines = '\n'.join(f'# {c!s}' for c in comment.comment.splitlines())
        self.print(f'\n{lines}\n')

    def walk_EOLComment(self, comment: grammars.EOLComment):
        self.walk_Comment(comment)

    def walk_EOF(self, eof: grammars.EOF):
        self.print('self._check_eof()')

    def walk_Group(self, group: grammars.Group):
        self.print('with self._group():')
        with self.indent():
            self.walk(group.exp)

    def walk_Token(self, token: grammars.Token):
        self.print(f'self._token({token.token!r})')

    def walk_Constant(self, constant: grammars.Constant):
        self.print(f'self._constant({constant.literal!r})')

    def walk_Alert(self, alert: grammars.Alert):
        self.print(f'self._alert({alert.literal!r}, {alert.level})')

    def walk_Pattern(self, pattern: grammars.Pattern):
        self.print(f'self._pattern({regexp(pattern.pattern)})')

    def walk_Lookahead(self, lookahead: grammars.Lookahead):
        self.print('with self._if():')
        with self.indent():
            self.walk(lookahead.exp)

    def walk_NegativeLookahead(self, lookahead: grammars.NegativeLookahead):
        self.print('with self._ifnot():')
        with self.indent():
            self.walk(lookahead.exp)

    def walk_Sequence(self, seq: grammars.Sequence):
        self.walk(seq.sequence)
        self._gen_defines_declaration(seq)

    def walk_Choice(self, choice: grammars.Choice):
        if len(choice.options) == 1:
            self.walk(choice.options[0])
            return

        firstset = choice.lookahead_str()
        if firstset:
            msglines = textwrap.wrap(firstset, width=40)
            error = ['expecting one of: ', *msglines]
        else:
            error = ['no available options']
        errors = repr(' '.join(error))
        raisestr = f'raise self.newexcept({errors}) from None'

        if not self.fitsfmt(errors, 3):
            errors = '\n'.join(repr(e) for e in error)
            one_liner = False
        elif self.fitsfmt(raisestr, 2):
            one_liner = True
        else:
            one_liner = False

        self.print('with self._choice():')
        with self.indent():
            self.walk(choice.options)
            self.print('if self._no_more_options:')
            with self.indent():
                if one_liner:
                    self.print(raisestr)
                else:
                    self.print('raise self.newexcept(')
                    with self.indent():
                        self.print(errors)
                    self.print(') from None')

    def walk_Option(self, option: grammars.Option):
        self.print('with self._option():')
        with self.indent():
            self.walk(option.exp)

    def walk_Optional(self, optional: grammars.Optional):
        self.print('with self._optional():')
        with self.indent():
            self.walk(optional.exp)

    def walk_EmptyClosure(self, closure: grammars.EmptyClosure):
        self.print('self._empty_closure()')

    def walk_Closure(self, closure: grammars.Closure):
        n = self._gen_block(closure.exp)
        self.print()
        self.print(f'self._closure(block{n})')

    def walk_PositiveClosure(self, closure: grammars.PositiveClosure):
        n = self._gen_block(closure.exp)
        self.print()
        self.print(f'self._positive_closure(block{n})')

    def walk_Join(self, join: grammars.Join):
        m = self._gen_block(join.sep, name='sep')
        n = self._gen_block(join.exp)
        self.print()
        self.print(f'self._join(block{n}, sep{m})')

    def walk_PositiveJoin(self, join: grammars.PositiveJoin):
        m = self._gen_block(join.sep, name='sep')
        n = self._gen_block(join.exp)
        self.print()
        self.print(f'self._positive_join(block{n}, sep{m})')

    def walk_LeftJoin(self, join: grammars.LeftJoin):
        m = self._gen_block(join.sep, name='sep')
        n = self._gen_block(join.exp)
        self.print()
        self.print(f'self._left_join(block{n}, sep{m})')

    def walk_RightJoin(self, join: grammars.RightJoin):
        m = self._gen_block(join.sep, name='sep')
        n = self._gen_block(join.exp)
        self.print()
        self.print(f'self._right_join(block{n}, sep{m})')

    def walk_Gather(self, gather: grammars.Gather):
        m = self._gen_block(gather.sep, name='sep')
        n = self._gen_block(gather.exp)
        self.print()
        self.print(f'self._gather(block{n}, sep{m})')

    def walk_PositiveGather(self, gather: grammars.PositiveGather):
        m = self._gen_block(gather.sep, name='sep')
        n = self._gen_block(gather.exp)
        self.print()
        self.print(f'self._positive_gather(block{n}, sep{m})')

    def walk_SkipTo(self, skipto: grammars.SkipTo):
        n = self._gen_block(skipto.exp)
        self.print()
        self.print(f'self._skip_to(block{n})')

    def walk_Named(self, named: grammars.Named):
        self.print(f"with self._setname('{named.name}'):")
        with self.indent():
            self.walk(named.exp)

    def walk_NamedList(self, named: grammars.Named):
        self.print(f"with self._addname('{named.name}'):")
        with self.indent():
            self.walk(named.exp)

    def walk_Override(self, override: grammars.Override):
        self.walk_Named(override)

    def walk_OverrideList(self, override: grammars.OverrideList):
        self.walk_NamedList(override)

    def _gen_keywords(self, grammar: grammars.Grammar):
        keywords = [str(k) for k in grammar.keywords if k is not None]
        if not keywords:
            self.print('KEYWORDS: set[str] = set()')
        else:
            self.print('KEYWORDS: set[str] = {')
            with self.indent():
                keywords_str = '\n'.join(f'    {k!r},' for k in sorted(keywords))
                self.print(keywords_str)
            self.print('}')

        self.print()
        self.print()

    def _gen_init(self, grammar: grammars.Grammar):
        start = grammar.config.start or grammar.rules[0].name

        whitespace = grammar.config.whitespace
        if whitespace is Undefined:  # the default value
            whitespace = None
        elif whitespace is not None:
            whitespace = regexp(whitespace)

        self.print(f'''
                config = ParserConfig.new(
                    config,
                    whitespace={whitespace},
                    nameguard={grammar.config.nameguard},
                    ignorecase={grammar.config.ignorecase},
                    namechars={grammar.config.namechars or ""!r},
                    parseinfo={grammar.config.parseinfo},
                    comments={regexp(grammar.config.comments)},
                    eol_comments={regexp(grammar.config.eol_comments)},
                    keywords=KEYWORDS,
                    start={start!r},
                )
                config = config.override(**settings)
            ''')
        self.print()

    def _gen_buffering(self, grammar: grammars.Grammar, parser_name: str):
        self.print(f'class {parser_name}Buffer(Buffer):')

        with self.indent():
            self.print(
                'def __init__(self, text, /, config: ParserConfig | None = None, **settings):',
            )
            with self.indent():
                self._gen_init(grammar)
                self.print('super().__init__(text, config=config)')
        self.print()

    def _gen_parsing(self, grammar: grammars.Grammar, parser_name: str):
        self.print()
        self.print(f'class {parser_name}Parser(Parser):')
        with self.indent():
            self.print(
                'def __init__(self, /, config: ParserConfig | None = None, **settings):',
            )
            with self.indent():
                self._gen_init(grammar)
                self.print('super().__init__(config=config)')
            self.print()
            self.walk(grammar.rules)
        self.print()

    def _gen_defines_declaration(self, node: grammars.Model):
        defines = compress_seq(node.defines())
        ldefs = {safe_name(d) for d, value in defines if value}
        sdefs = {safe_name(d) for d, value in defines if not value and d not in ldefs}

        if not (sdefs or ldefs):
            return

        sdefs_str = ', '.join(sorted(repr(d) for d in sdefs))
        ldefs_str = ', '.join(sorted(repr(d) for d in ldefs))

        definestr = f'self._define([{sdefs_str}], [{ldefs_str}])'
        if not ldefs or self.fitsfmt(definestr, 1):
            self.print(definestr)
        else:
            self.print('self._define(')
            with self.indent():
                self.print(f'[{sdefs_str}],')
                self.print(f'[{ldefs_str}],')
            self.print(')')

    def _gen_block(self, exp: grammars.Model, name='block'):
        if () in exp.lookahead():
            raise CodegenError(
                f'{exp!r} may repeat empty sequence @{exp.line} {exp.lookahead()!r}',
            )

        n = self._next_n()
        self.print()
        self.print(f'def {name}{n}():')
        with self.indent():
            self.walk(exp)

        return n



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any

from .. import grammars
from ..util.abctools import join_lists
from ..util.string import regexp
from ..util.string import unicode_display_len as ulen
from ..walkers import NodeWalker
from .railmath import (
    ETX,
    Rails,
    assert_one_length,
    lay_out,
    loop,
    stopnloop,
    weld,
)


def tracks(model: grammars.Grammar):
    walker = RailroadNodeWalker()
    return walker.walk(model)


def text(model: grammars.Grammar) -> str:
    return '\n'.join(line.rstrip() for line in tracks(model))


def draw(model: grammars.Grammar):
    for line in tracks(model):
        print(line.rstrip())


class RailroadNodeWalker(NodeWalker):
    def __init__(self):
        super().__init__()

    def walk(self, node: Any, *args, **kwargs) -> Any:
        return list(super().walk(node))

    def walk_default(self, node: grammars.Model) -> Rails:
        return [f' <{node!r}> ']

    def walk_decorator(self, decorator: grammars.Decorator) -> Rails:
        return self.walk(decorator.exp)

    def walk_grammar(self, grammar: grammars.Grammar) -> Rails:
        return join_lists(*(self.walk(r) for r in grammar.rules))

    def walk_rule(self, rule: grammars.Rule) -> Rails:
        decorators = ''
        if rule.decorators:
            decorators = ' '.join(f'@{d}' for d in rule.decorators) + ' '

        params = ''
        if rule.params:
            params = '∷' + ','.join(p for p in rule.params)
        if rule.kwparams:
            params = ',' + ','.join(f'{k}={v}' for k, v in rule.kwparams.items())

        leftrec = ''
        if rule.is_leftrec:
            leftrec = '⟳'
        elif not rule.is_memoizable:
            leftrec = '⊬'

        base = ''
        if rule.base:
            base = f"≤({rule.base})"

        out = [f'{decorators}{leftrec}{rule.name}{base}{params} ●─']
        out = weld(out, self.walk(rule.exp))
        if ETX not in out:
            out = weld(out, ['─■'])
        out += [' ' * ulen(out[0])]
        return assert_one_length(out)

    def walk_optional(self, optional: grammars.Optional) -> Rails:
        out = weld(['→'], self.walk(optional.exp))
        return lay_out([out, ['→']])

    def walk_closure(self, closure: grammars.Closure) -> Rails:
        return loop(self.walk_decorator(closure))

    def walk_positive_closure(self, closure: grammars.Closure) -> Rails:
        return stopnloop(self.walk_decorator(closure))

    def walk_join(self, join: grammars.Join) -> Rails:
        sep = weld(self.walk(join.sep), [' ✂ ─'])
        out = weld(sep, self.walk(join.exp))
        return loop(out)

    def walk_positive_join(self, join: grammars.PositiveJoin) -> Rails:
        sep = weld(self.walk(join.sep), [' ✂ ─'])
        out = weld(sep, self.walk(join.exp))
        return stopnloop(out)

    def walk_choice(self, choice: grammars.Choice) -> Rails:
        return lay_out([self.walk(o) for o in choice.options])

    def walk_option(self, option: grammars.Option) -> Rails:
        return self.walk(option.exp)

    def walk_sequence(self, s: grammars.Sequence) -> Rails:
        return weld(*(self.walk(e) for e in s.sequence))

    def walk_call(self, call: grammars.Call) -> Rails:
        return [f"{call.name}"]

    def walk_pattern(self, pattern: grammars.Pattern) -> Rails:
        pat = regexp(pattern.pattern).replace("r'", "").rstrip("'")
        return [f"/{pat}/─"]

    def walk_token(self, token: grammars.Token) -> Rails:
        return [f"{token.token!r}"]

    def walk_eof(self, eof: grammars.EOF) -> Rails:
        return [f"⇥ {ETX} "]

    def walk_lookahead(self, la: grammars.Lookahead) -> Rails:
        out = weld(['─ &['], self.walk(la.exp))
        out = weld(out, [']'])
        return out

    def walk_negative_lookahead(self, la: grammars.NegativeLookahead) -> Rails:
        out = weld(['─ !['], self.walk(la.exp))
        out = weld(out, [']'])
        return out

    def walk_void(self, v: grammars.Void) -> Rails:
        return [" ∅ "]

    def walk_cut(self, cut: grammars.Cut) -> Rails:
        return [" ✂ ─"]

    def walk_fail(self, v) -> Rails:
        return [" ⚠ "]

    def walk_constant(self, constant: grammars.Constant) -> Rails:
        return [f'`{constant.literal}`']

    def walk_dot(self, dot: grammars.Dot):
        return [" ∀ "]

    def walk_group(self, group: grammars.Group):
        return self.walk_decorator(group)

    def walk_alert(self, alert: grammars.Alert):
        return [f'{'^' * alert.level}`{alert.literal}`']

    def walk_skip_to(self, skipto: grammars.SkipTo):
        return weld([' ->('], self.walk(skipto.exp), [')'])

    def walk_rule_include(self, include: grammars.RuleInclude):
        return [f' >({include.rule.name}) ']

    def walk_based_rule(self, rule: grammars.BasedRule):
        out = [f'{rule.name} < {rule.baserule}●─']
        out = weld(out, self.walk(rule.rhs))
        if ETX not in out:
            out = weld(out, ['─■'])
        out += [' ' * ulen(out[0])]
        return assert_one_length(out)

    def walk_named(self, named: grammars.Named) -> Rails:
        return weld([f' [{named.name}]('], self.walk(named.exp), [')'])

    def walk_named_list(self, named: grammars.NamedList) -> Rails:
        return weld([f' [`{named.name}`]+('], self.walk(named.exp), [')'])

    def walk_override(self, override: grammars.Override) -> Rails:
        return weld([' @('], self.walk(override.exp), [')'])

    def walk_override_list(self, override: grammars.OverrideList):
        return weld([' @+('], self.walk(override.exp), [')'])



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import argparse
import importlib
import sys
from pathlib import Path

from .. import railroads
from .._version import __version__
from ..exceptions import ParseException
from ..ngcodegen import modelgen, pythongen
from ..parserconfig import ParserConfig
from ..util import eval_escapes
from . import api

__all__ = ['tatsu_main']


DESCRIPTION = (
    '竜TatSu takes a grammar in extended EBNF'
    ' as input, and outputs a memoizing'
    ' PEG/Packrat parser in Python.'
)


def parse_args():
    argparser = argparse.ArgumentParser(
        prog='tatsu',
        description=DESCRIPTION,
        add_help=False,
    )

    main_mode = argparser.add_mutually_exclusive_group()
    main_mode.add_argument(
        '--generate-parser',
        help='generate parser code from the grammar (default)',
        action='store_true',
    )
    main_mode.add_argument(
        '--draw',
        '-d',
        help=(
            'generate a diagram of the grammar'
            ' (.svg, .png, .jpeg, .dot, ...'
            ' / requres --outfile)'
        ),
        action='store_true',
    )
    main_mode.add_argument(
        '--railroad',
        '-r',
        help='output a railroad diagram of the grammar in ASCII/Text Art',
        action='store_true',
    )
    main_mode.add_argument(
        '--object-model',
        '-g',
        help='generate object model from the class names given as rule arguments',
        action='store_true',
    )
    main_mode.add_argument(
        '--pretty',
        '-p',
        help='generate a prettified version of the input grammar',
        action='store_true',
    )
    main_mode.add_argument(
        '--pretty-lean',
        help='like --pretty, but without name: or ::Parameter annotations',
        action='store_true',
    )

    ebnf_opts = argparser.add_argument_group('parse-time options')
    argparser.add_argument(
        'filename',
        metavar='GRAMMAR',
        help='the filename of the TatSu grammar to parse',
    )
    ebnf_opts.add_argument(
        '--color',
        '-c',
        help='use color in traces (requires the colorama library)',
        action='store_true',
    )
    ebnf_opts.add_argument(
        '--trace',
        '-t',
        help='produce verbose parsing output',
        action='store_true',
    )

    generation_opts = argparser.add_argument_group('generation options')
    generation_opts.add_argument(
        '--left-recursion',
        '-l',
        help='turns left-recursion support on',
        dest='left_recursion',
        action='store_true',
        default=None,
    )
    generation_opts.add_argument(
        '--name',
        '-m',
        metavar='NAME',
        help='Name for the grammar (defaults to GRAMMAR base name)',
    )
    generation_opts.add_argument(
        '--nameguard',
        '-n',
        help='allow tokens that are prefixes of others',
        dest='nameguard',
        action='store_false',
        default=None,  # None allows grammar specified
    )
    generation_opts.add_argument(
        '--outfile',
        '--output',
        '-o',
        metavar='FILE',
        help='output file (default is stdout)',
    )
    generation_opts.add_argument(
        '--object-model-outfile',
        '-G',
        metavar='FILE',
        help='generate object model and save to FILE',
    )
    generation_opts.add_argument(
        '--whitespace',
        '-w',
        metavar='CHARACTERS',
        help='characters to skip during parsing (use "" to disable)',
    )

    def import_class(path):
        try:
            spath = path.rsplit('.', 1)
            module = importlib.import_module(spath[0])

            return getattr(module, spath[1])
        except Exception as e:
            raise argparse.ArgumentTypeError(f"Couldn't find class {path}") from e

    generation_opts.add_argument(
        '--base-type',
        metavar='CLASSPATH',
        help='class to use as base type for the object model, for example "mymodule.MyNode"',
        type=import_class,
    )

    std_args = argparser.add_argument_group('common options')
    std_args.add_argument(
        '--help',
        '-h',
        help='show this help message and exit',
        action='help',
    )
    std_args.add_argument(
        '--version',
        '-V',
        help='provide version information and exit',
        action='version',
        version=f'TatSu {__version__}',
    )

    args = argparser.parse_args()

    if args.draw and not args.outfile:
        argparser.error('--draw requires --outfile')

    return args


def prepare_for_output(filename: str):
    if not filename:
        return

    path = Path(filename)

    dirname = path.parent
    if not dirname.exists():
        dirname.mkdir(parents=True, exist_ok=True)

    if path.is_file():
        path.unlink()


def save(filename: str, content: str):
    Path(filename).write_text(content, encoding='utf-8')


def tatsu_main():
    args = parse_args()

    if args.whitespace:
        args.whitespace = eval_escapes(args.whitespace)

    outfile = args.outfile
    prepare_for_output(outfile)
    prepare_for_output(args.object_model_outfile)

    grammar = Path(args.filename).read_text()

    try:
        config = ParserConfig(
            trace=args.trace,
            filename=args.filename,
            colorize=args.color,
            left_recursion=args.left_recursion,
            nameguard=args.nameguard,
            whitespace=str(args.whitespace),
        )
        model = api.compile(grammar, args.name, asmodel=True, config=config)

        if args.draw:
            from .. import diagrams

            diagrams.draw(outfile, model)
        elif args.railroad:
            railroad = railroads.text(model)
            if outfile:
                save(outfile, railroad)
            else:
                print(railroad)
        else:
            if args.pretty:
                result = model.pretty()
            elif args.pretty_lean:
                result = model.pretty_lean()
            elif args.object_model:
                result = modelgen(model, basetype=args.base_type)
            else:
                result = pythongen(model)

            if outfile:
                save(outfile, result)
            else:
                print(result)

            # if requested, always save it
            if args.object_model_outfile:
                save(
                    args.object_model_outfile,
                    modelgen(model, basetype=args.base_type),
                )

        print('─' * 72, file=sys.stderr)
        print(f'{len(grammar.split()):12,d}  lines in grammar', file=sys.stderr)
        print(f'{len(model.rules):12,d}  rules in grammar', file=sys.stderr)
        print(f'{model.nodecount():12,d}  nodes in AST', file=sys.stderr)
    except ParseException as e:
        print(e, file=sys.stderr)
        sys.exit(1)



import pytest

from tatsu import diagrams
from tatsu.tool import compile


@pytest.mark.skipif(not diagrams.available(), reason='graphviz is not available')
def test_dot():
    grammar = """
        start = "foo\\nbar" $;
    """
    m = compile(grammar, 'Diagram')
    diagrams.draw('tmp/diagram.svg', m)
